# Copyright (c) 2017 DataCore Software Corp. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Unit tests for classes that are used to invoke DataCore SANsymphony API."""

from unittest import mock

from oslo_utils import units
import suds
from suds.sax import parser
from suds import wsdl

from cinder.tests.unit import test
from cinder.volume.drivers.datacore import api
from cinder.volume.drivers.datacore import exception


class FakeWebSocketException(Exception):
    pass


class DataCoreClientTestCase(test.TestCase):
    """Tests for the DataCore SANsymphony client."""

    def setUp(self):
        super(DataCoreClientTestCase, self).setUp()
        self.mock_storage_services = mock.MagicMock()
        self.mock_executive_service = mock.MagicMock()

        self.mock_suds_client = mock.MagicMock()
        self.mock_object(
            api.suds_client, 'Client', return_value=self.mock_suds_client)

        self.mock_channel = mock.MagicMock()
        mock_websocket = self.mock_object(api, 'websocket')
        mock_websocket.WebSocketException = FakeWebSocketException
        mock_websocket.create_connection.return_value = self.mock_channel

        setattr(self.mock_suds_client.service.__getitem__,
                'side_effect',
                self._get_service_side_effect)

        self.client = api.DataCoreClient('hostname', 'username', 'password', 1)
        self.client.API_RETRY_INTERVAL = 0

        # Make sure failure logging does not get emitted during testing
        self.mock_object(api, 'LOG')

    def _get_service_side_effect(self, service_name):
        self.assertIn(service_name,
                      [
                          api.DataCoreClient.STORAGE_SERVICES_BINDING,
                          api.DataCoreClient.EXECUTIVE_SERVICE_BINDING
                      ])

        if service_name is api.DataCoreClient.STORAGE_SERVICES_BINDING:
            return self.mock_storage_services
        else:
            return self.mock_executive_service

    def _assert_storage_services_method_called(self, method_name):
        return self.mock_storage_services.__getitem__.assert_called_with(
            method_name)

    @property
    def mock_storage_service_context(self):
        return self.mock_storage_services.__getitem__()()

    @property
    def mock_executive_service_context(self):
        return self.mock_executive_service.__getitem__()()

    def test_process_request_failed(self):
        def fail_with_socket_error():
            raise FakeWebSocketException()

        def fail_with_web_fault(message):
            fault = mock.Mock()
            fault.faultstring = "General error."
            document = mock.Mock()
            raise suds.WebFault(fault, document)

        self.mock_channel.recv.side_effect = fail_with_socket_error
        self.assertRaises(exception.DataCoreConnectionException,
                          self.client.get_server_groups)
        self.mock_channel.recv.side_effect = None

        (self.mock_storage_service_context.process_reply
         .side_effect) = fail_with_web_fault
        self.assertRaises(exception.DataCoreFaultException,
                          self.client.get_server_groups)

    def test_channel_closing_failed(self):
        def fail_with_socket_error():
            raise FakeWebSocketException()

        def fail_with_web_fault(message):
            fault = mock.Mock()
            fault.faultstring = "General error."
            document = mock.Mock()
            raise suds.WebFault(fault, document)

        self.mock_channel.close.side_effect = fail_with_socket_error
        (self.mock_storage_service_context.process_reply
         .side_effect) = fail_with_web_fault
        self.assertRaises(exception.DataCoreFaultException,
                          self.client.get_server_groups)

    def test_update_api_endpoints(self):
        def fail_with_socket_error():
            try:
                raise FakeWebSocketException()
            finally:
                self.mock_channel.recv.side_effect = None

        self.mock_channel.recv.side_effect = fail_with_socket_error

        mock_executive_endpoints = [{
            'network_address': '127.0.0.1:3794',
            'http_endpoint': 'http://127.0.0.1:3794/',
            'ws_endpoint': 'ws://127.0.0.1:3794/',
        }]
        self.mock_object(self.client,
                         '_executive_service_endpoints',
                         mock_executive_endpoints)

        mock_storage_endpoint = {
            'network_address': '127.0.0.1:3794',
            'http_endpoint': 'http://127.0.0.1:3794/',
            'ws_endpoint': 'ws://127.0.0.1:3794/',
        }
        self.mock_object(self.client,
                         '_storage_services_endpoint',
                         mock_storage_endpoint)

        node = mock.Mock()
        node.HostAddress = '127.0.0.1:3794'
        reply = mock.MagicMock()
        reply.RegionNodeData = [node]
        self.mock_storage_service_context.process_reply.return_value = reply

        result = self.client.get_server_groups()
        self.assertIsNotNone(result)

    def test_update_api_endpoints_failed(self):
        def fail_with_socket_error():
            try:
                raise FakeWebSocketException()
            finally:
                self.mock_channel.recv.side_effect = None

        self.mock_channel.recv.side_effect = fail_with_socket_error

        mock_executive_endpoints = [{
            'network_address': '127.0.0.1:3794',
            'http_endpoint': 'http://127.0.0.1:3794/',
            'ws_endpoint': 'ws://127.0.0.1:3794/',
        }]
        self.mock_object(self.client,
                         '_executive_service_endpoints',
                         mock_executive_endpoints)

        reply = mock.MagicMock()
        reply.RegionNodeData = []
        self.mock_storage_service_context.process_reply.return_value = reply

        self.mock_executive_service_context.process_reply.return_value = None

        result = self.client.get_server_groups()
        self.assertIsNotNone(result)

    def test_get_server_groups(self):
        self.client.get_server_groups()
        self._assert_storage_services_method_called('GetServerGroups')

    def test_get_servers(self):
        self.client.get_servers()
        self._assert_storage_services_method_called('GetServers')

    def test_get_disk_pools(self):
        self.client.get_disk_pools()
        self._assert_storage_services_method_called('GetDiskPools')

    def test_get_logical_disks(self):
        self.client.get_logical_disks()
        self._assert_storage_services_method_called('GetLogicalDisks')

    def test_create_pool_logical_disk(self):
        pool_id = 'pool_id'
        pool_volume_type = 'Striped'
        size = 1 * units.Gi
        min_quota = 1
        max_quota = 1 * units.Gi
        self.client.create_pool_logical_disk(
            pool_id, pool_volume_type, size, min_quota, max_quota)
        self._assert_storage_services_method_called('CreatePoolLogicalDisk')

    def test_delete_logical_disk(self):
        logical_disk_id = 'disk_id'
        self.client.delete_logical_disk(logical_disk_id)
        self._assert_storage_services_method_called('DeleteLogicalDisk')

    def test_get_logical_disk_chunk_allocation_map(self):
        logical_disk_id = 'disk_id'
        self.client.get_logical_disk_chunk_allocation_map(logical_disk_id)
        self._assert_storage_services_method_called(
            'GetLogicalDiskChunkAllocationMap')

    def test_get_next_virtual_disk_alias(self):
        base_alias = 'volume'
        self.client.get_next_virtual_disk_alias(base_alias)
        self._assert_storage_services_method_called('GetNextVirtualDiskAlias')

    def test_get_virtual_disks(self):
        self.client.get_virtual_disks()
        self._assert_storage_services_method_called('GetVirtualDisks')

    def test_build_virtual_disk_data(self):
        disk_alias = 'alias'
        disk_type = 'Mirrored'
        size = 1 * units.Gi
        description = 'description'
        storage_profile_id = 'storage_profile_id'

        vd_data = self.client.build_virtual_disk_data(
            disk_alias, disk_type, size, description, storage_profile_id)

        self.assertEqual(disk_alias, vd_data.Alias)
        self.assertEqual(size, vd_data.Size.Value)
        self.assertEqual(description, vd_data.Description)
        self.assertEqual(storage_profile_id, vd_data.StorageProfileId)
        self.assertTrue(hasattr(vd_data, 'Type'))
        self.assertTrue(hasattr(vd_data, 'SubType'))
        self.assertTrue(hasattr(vd_data, 'DiskStatus'))
        self.assertTrue(hasattr(vd_data, 'RecoveryPriority'))

    def test_create_virtual_disk_ex2(self):
        disk_alias = 'alias'
        disk_type = 'Mirrored'
        size = 1 * units.Gi
        description = 'description'
        storage_profile_id = 'storage_profile_id'
        first_disk_id = 'disk_id'
        second_disk_id = 'disk_id'
        add_redundancy = True
        vd_data = self.client.build_virtual_disk_data(
            disk_alias, disk_type, size, description, storage_profile_id)
        self.client.create_virtual_disk_ex2(
            vd_data, first_disk_id, second_disk_id, add_redundancy)
        self._assert_storage_services_method_called('CreateVirtualDiskEx2')

    def test_set_virtual_disk_size(self):
        disk_id = 'disk_id'
        size = 1 * units.Gi
        self.client.set_virtual_disk_size(disk_id, size)
        self._assert_storage_services_method_called('SetVirtualDiskSize')

    def test_delete_virtual_disk(self):
        virtual_disk_id = 'disk_id'
        delete_logical_disks = True
        self.client.delete_virtual_disk(virtual_disk_id, delete_logical_disks)
        self._assert_storage_services_method_called('DeleteVirtualDisk')

    def test_serve_virtual_disks_to_host(self):
        host_id = 'host_id'
        disks = ['disk_id']
        self.client.serve_virtual_disks_to_host(host_id, disks)
        self._assert_storage_services_method_called('ServeVirtualDisksToHost')

    def test_unserve_virtual_disks_from_host(self):
        host_id = 'host_id'
        disks = ['disk_id']
        self.client.unserve_virtual_disks_from_host(host_id, disks)
        self._assert_storage_services_method_called(
            'UnserveVirtualDisksFromHost')

    def test_unserve_virtual_disks_from_port(self):
        port_id = 'port_id'
        disks = ['disk_id']
        self.client.unserve_virtual_disks_from_port(port_id, disks)
        self._assert_storage_services_method_called(
            'UnserveVirtualDisksFromPort')

    def test_bind_logical_disk(self):
        disk_id = 'disk_id'
        logical_disk_id = 'disk_id'
        role = 'Second'
        create_mirror_mappings = True
        create_client_mappings = False
        add_redundancy = True
        self.client.bind_logical_disk(
            disk_id, logical_disk_id, role, create_mirror_mappings,
            create_client_mappings, add_redundancy)
        self._assert_storage_services_method_called(
            'BindLogicalDisk')

    def test_get_snapshots(self):
        self.client.get_snapshots()
        self._assert_storage_services_method_called('GetSnapshots')

    def test_create_snapshot(self):
        disk_id = 'disk_id'
        name = 'name'
        description = 'description'
        pool_id = 'pool_id'
        snapshot_type = 'Full'
        duplicate_disk_id = False
        storage_profile_id = 'profile_id'
        self.client.create_snapshot(
            disk_id, name, description, pool_id, snapshot_type,
            duplicate_disk_id, storage_profile_id)
        self._assert_storage_services_method_called('CreateSnapshot')

    def test_delete_snapshot(self):
        snapshot_id = "snapshot_id"
        self.client.delete_snapshot(snapshot_id)
        self._assert_storage_services_method_called('DeleteSnapshot')

    def test_get_storage_profiles(self):
        self.client.get_storage_profiles()
        self._assert_storage_services_method_called('GetStorageProfiles')

    def test_designate_map_store(self):
        pool_id = 'pool_id'
        self.client.designate_map_store(pool_id)
        self._assert_storage_services_method_called('DesignateMapStore')

    def test_get_performance_by_type(self):
        types = ['DiskPoolPerformance']
        self.client.get_performance_by_type(types)
        self._assert_storage_services_method_called('GetPerformanceByType')

    def test_get_ports(self):
        self.client.get_ports()
        self._assert_storage_services_method_called('GetPorts')

    def test_build_scsi_port_data(self):
        host_id = 'host_id'
        port_name = 'port_name'
        port_mode = 'Initiator'
        port_type = 'iSCSI'

        port_data = self.client.build_scsi_port_data(
            host_id, port_name, port_mode, port_type)

        self.assertEqual(host_id, port_data.HostId)
        self.assertEqual(port_name, port_data.PortName)
        self.assertTrue(hasattr(port_data, 'PortMode'))
        self.assertTrue(hasattr(port_data, 'PortType'))

    def test_register_port(self):
        port_data = self.client.build_scsi_port_data(
            'host_id', 'port_name', 'initiator', 'iSCSI')
        self.client.register_port(port_data)
        self._assert_storage_services_method_called('RegisterPort')

    def test_assign_port(self):
        client_id = 'client_id'
        port_id = 'port_id'
        self.client.assign_port(client_id, port_id)
        self._assert_storage_services_method_called('AssignPort')

    def test_set_server_port_properties(self):
        port_id = 'port_id'
        port_properties = mock.MagicMock()
        self.client.set_server_port_properties(port_id, port_properties)
        self._assert_storage_services_method_called('SetServerPortProperties')

    def test_build_access_token(self):
        initiator_node_name = 'initiator'
        initiator_username = 'initiator_username'
        initiator_password = 'initiator_password'
        mutual_authentication = True
        target_username = 'target_username'
        target_password = 'target_password'

        access_token = self.client.build_access_token(
            initiator_node_name, initiator_username, initiator_password,
            mutual_authentication, target_username, target_password)

        self.assertEqual(initiator_node_name, access_token.InitiatorNodeName)
        self.assertEqual(initiator_username, access_token.InitiatorUsername)
        self.assertEqual(initiator_password, access_token.InitiatorPassword)
        self.assertEqual(mutual_authentication,
                         access_token.MutualAuthentication)
        self.assertEqual(target_username, access_token.TargetUsername)
        self.assertEqual(target_password, access_token.TargetPassword)

    def test_set_access_token(self):
        port_id = 'port_id'
        access_token = self.client.build_access_token(
            'initiator_name', None, None, False, 'initiator_name', 'password')
        self.client.set_access_token(port_id, access_token)
        self._assert_storage_services_method_called('SetAccessToken')

    def test_get_clients(self):
        self.client.get_clients()
        self._assert_storage_services_method_called('GetClients')

    def test_register_client(self):
        host_name = 'name'
        description = 'description'
        machine_type = 'Other'
        mode = 'PreferredServer'
        preferred_server_ids = None
        self.client.register_client(
            host_name, description, machine_type, mode, preferred_server_ids)
        self._assert_storage_services_method_called('RegisterClient')

    def test_set_client_capabilities(self):
        client_id = 'client_id'
        mpio = True
        alua = True
        self.client.set_client_capabilities(client_id, mpio, alua)
        self._assert_storage_services_method_called('SetClientCapabilities')

    def test_get_target_domains(self):
        self.client.get_target_domains()
        self._assert_storage_services_method_called('GetTargetDomains')

    def test_create_target_domain(self):
        initiator_host_id = 'host_id'
        target_host_id = 'host_id'
        self.client.create_target_domain(initiator_host_id, target_host_id)
        self._assert_storage_services_method_called('CreateTargetDomain')

    def test_delete_target_domain(self):
        domain_id = 'domain_id'
        self.client.delete_target_domain(domain_id)
        self._assert_storage_services_method_called('DeleteTargetDomain')

    def test_get_target_devices(self):
        self.client.get_target_devices()
        self._assert_storage_services_method_called('GetTargetDevices')

    def test_build_scsi_port_nexus_data(self):
        initiator_id = 'initiator_id'
        target_id = 'target_id'

        nexus = self.client.build_scsi_port_nexus_data(initiator_id, target_id)

        self.assertEqual(initiator_id, nexus.InitiatorPortId)
        self.assertEqual(target_id, nexus.TargetPortId)

    def test_create_target_device(self):
        domain_id = 'domain_id'
        nexus = self.client.build_scsi_port_nexus_data('initiator_id',
                                                       'target_id')
        self.client.create_target_device(domain_id, nexus)
        self._assert_storage_services_method_called('CreateTargetDevice')

    def test_delete_target_device(self):
        device_id = 'device_id'
        self.client.delete_target_device(device_id)
        self._assert_storage_services_method_called('DeleteTargetDevice')

    def test_get_next_free_lun(self):
        device_id = 'device_id'
        self.client.get_next_free_lun(device_id)
        self._assert_storage_services_method_called('GetNextFreeLun')

    def test_get_logical_units(self):
        self.client.get_logical_units()
        self._assert_storage_services_method_called('GetLogicalUnits')

    def test_map_logical_disk(self):
        disk_id = 'disk_id'
        lun = 0
        host_id = 'host_id'
        mapping_type = 'Client'
        initiator_id = 'initiator_id'
        target_id = 'target_id'
        nexus = self.client.build_scsi_port_nexus_data(initiator_id, target_id)
        self.client.map_logical_disk(
            disk_id, nexus, lun, host_id, mapping_type)
        self._assert_storage_services_method_called('MapLogicalDisk')

    def test_unmap_logical_disk(self):
        logical_disk_id = 'disk_id'
        nexus = self.client.build_scsi_port_nexus_data('initiator_id',
                                                       'target_id')
        self.client.unmap_logical_disk(logical_disk_id, nexus)
        self._assert_storage_services_method_called('UnmapLogicalDisk')


FAKE_WSDL_DOCUMENT = """<?xml version="1.0" encoding="utf-8"?>
<wsdl:definitions name="ExecutiveServices"
                  targetNamespace="http://tempuri.org/"
                  xmlns:wsdl="http://schemas.xmlsoap.org/wsdl/"
                  xmlns:soap="http://schemas.xmlsoap.org/wsdl/soap/"
                  xmlns:tns="http://tempuri.org/"
                  xmlns:wsa10="http://www.w3.org/2005/08/addressing"
                  xmlns:wsaw="http://www.w3.org/2006/05/addressing/wsdl">
    <wsdl:types>
        <xs:schema elementFormDefault="qualified"
                   targetNamespace="http://tempuri.org/"
                   xmlns:xs="http://www.w3.org/2001/XMLSchema">
            <xs:import
namespace="http://schemas.microsoft.com/2003/10/Serialization/Arrays"/>
            <xs:import
namespace="http://schemas.datacontract.org/2004/07/DataCore.Executive"/>
            <xs:element name="StartExecutive">
                <xs:complexType>
                    <xs:sequence/>
                </xs:complexType>
            </xs:element>
            <xs:element name="StartExecutiveResponse">
                <xs:complexType>
                    <xs:sequence/>
                </xs:complexType>
            </xs:element>
            <xs:element name="StopExecutive">
                <xs:complexType>
                    <xs:sequence/>
                </xs:complexType>
            </xs:element>
            <xs:element name="StopExecutiveResponse">
                <xs:complexType>
                    <xs:sequence/>
                </xs:complexType>
            </xs:element>
            <xs:element name="ExecutiveStarted">
                <xs:complexType>
                    <xs:sequence/>
                </xs:complexType>
            </xs:element>
            <xs:element name="ExecutiveStopped">
                <xs:complexType>
                    <xs:sequence/>
                </xs:complexType>
            </xs:element>
        </xs:schema>
    </wsdl:types>
    <wsdl:message name="IExecutiveServiceEx_StartExecutive_InputMessage">
        <wsdl:part name="parameters" element="tns:StartExecutive"/>
    </wsdl:message>
    <wsdl:message name="IExecutiveServiceEx_StartExecutive_OutputMessage">
        <wsdl:part name="parameters" element="tns:StartExecutiveResponse"/>
    </wsdl:message>
    <wsdl:message
name="IExecutiveServiceEx_StartExecutive_ExecutiveError_FaultMessage">
        <wsdl:part name="detail" element="ExecutiveError"/>
    </wsdl:message>
    <wsdl:message name="IExecutiveServiceEx_StopExecutive_InputMessage">
        <wsdl:part name="parameters" element="tns:StopExecutive"/>
    </wsdl:message>
    <wsdl:message name="IExecutiveServiceEx_StopExecutive_OutputMessage">
        <wsdl:part name="parameters" element="tns:StopExecutiveResponse"/>
    </wsdl:message>
    <wsdl:message
name="IExecutiveServiceEx_StopExecutive_ExecutiveError_FaultMessage">
        <wsdl:part name="detail" element="ExecutiveError"/>
    </wsdl:message>
    <wsdl:message
            name="IExecutiveServiceEx_ExecutiveStarted_OutputCallbackMessage">
        <wsdl:part name="parameters" element="tns:ExecutiveStarted"/>
    </wsdl:message>
    <wsdl:message
name="IExecutiveServiceEx_ExecutiveStopped_OutputCallbackMessage">
        <wsdl:part name="parameters" element="tns:ExecutiveStopped"/>
    </wsdl:message>
    <wsdl:portType name="IExecutiveServiceEx">
        <wsdl:operation name="StartExecutive">
            <wsdl:input
wsaw:Action="http://tempuri.org/IExecutiveService/StartExecutive"
message="tns:IExecutiveServiceEx_StartExecutive_InputMessage"/>
            <wsdl:output
wsaw:Action="http://tempuri.org/IExecutiveService/StartExecutiveResponse"
message="tns:IExecutiveServiceEx_StartExecutive_OutputMessage"/>
            <wsdl:fault wsaw:Action="ExecutiveError" name="ExecutiveError"
message="tns:IExecutiveServiceEx_StartExecutive_ExecutiveError_FaultMessage"/>
        </wsdl:operation>
        <wsdl:operation name="StopExecutive">
            <wsdl:input
wsaw:Action="http://tempuri.org/IExecutiveService/StopExecutive"
message="tns:IExecutiveServiceEx_StopExecutive_InputMessage"/>
            <wsdl:output
wsaw:Action="http://tempuri.org/IExecutiveService/StopExecutiveResponse"
message="tns:IExecutiveServiceEx_StopExecutive_OutputMessage"/>
            <wsdl:fault wsaw:Action="ExecutiveError" name="ExecutiveError"
message="tns:IExecutiveServiceEx_StopExecutive_ExecutiveError_FaultMessage"/>
        </wsdl:operation>
        <wsdl:operation name="ExecutiveStarted">
            <wsdl:output
wsaw:Action="http://tempuri.org/IExecutiveService/ExecutiveStarted"
message="tns:IExecutiveServiceEx_ExecutiveStarted_OutputCallbackMessage"/>
            <wsdl:fault wsaw:Action="ExecutiveError" name="ExecutiveError"
                        message="tns:"/>
        </wsdl:operation>
        <wsdl:operation name="ExecutiveStopped">
            <wsdl:output
wsaw:Action="http://tempuri.org/IExecutiveService/ExecutiveStopped"
message="tns:IExecutiveServiceEx_ExecutiveStopped_OutputCallbackMessage"/>
            <wsdl:fault wsaw:Action="ExecutiveError" name="ExecutiveError"
                        message="tns:"/>
        </wsdl:operation>
    </wsdl:portType>
    <wsdl:binding name="CustomBinding_IExecutiveServiceEx"
                  type="tns:IExecutiveServiceEx">
        <soap:binding transport="http://schemas.microsoft.com/soap/websocket"/>
        <wsdl:operation name="StartExecutive">
            <soap:operation
soapAction="http://tempuri.org/IExecutiveService/StartExecutive"
                    style="document"/>
            <wsdl:input>
                <soap:body use="literal"/>
            </wsdl:input>
            <wsdl:output>
                <soap:body use="literal"/>
            </wsdl:output>
            <wsdl:fault name="ExecutiveError">
                <soap:fault use="literal" name="ExecutiveError" namespace=""/>
            </wsdl:fault>
        </wsdl:operation>
        <wsdl:operation name="StopExecutive">
            <soap:operation
soapAction="http://tempuri.org/IExecutiveService/StopExecutive"
                    style="document"/>
            <wsdl:input>
                <soap:body use="literal"/>
            </wsdl:input>
            <wsdl:output>
                <soap:body use="literal"/>
            </wsdl:output>
            <wsdl:fault name="ExecutiveError">
                <soap:fault use="literal" name="ExecutiveError" namespace=""/>
            </wsdl:fault>
        </wsdl:operation>
        <wsdl:operation name="ExecutiveStarted">
            <soap:operation
soapAction="http://tempuri.org/IExecutiveService/ExecutiveStarted"
                    style="document"/>
            <wsdl:output>
                <soap:body use="literal"/>
            </wsdl:output>
            <wsdl:fault name="ExecutiveError">
                <soap:fault use="literal" name="ExecutiveError" namespace=""/>
            </wsdl:fault>
        </wsdl:operation>
        <wsdl:operation name="ExecutiveStopped">
            <soap:operation
soapAction="http://tempuri.org/IExecutiveService/ExecutiveStopped"
                    style="document"/>
            <wsdl:output>
                <soap:body use="literal"/>
            </wsdl:output>
            <wsdl:fault name="ExecutiveError">
                <soap:fault use="literal" name="ExecutiveError" namespace=""/>
            </wsdl:fault>
        </wsdl:operation>
    </wsdl:binding>
    <wsdl:service name="ExecutiveServices">
        <wsdl:port name="CustomBinding_IExecutiveServiceEx"
                   binding="tns:CustomBinding_IExecutiveServiceEx">
            <soap:address
                    location="ws://mns-vsp-001:3794/IExecutiveServiceEx"/>
            <wsa10:EndpointReference>
                <wsa10:Address>ws://mns-vsp-001:3794/IExecutiveServiceEx
                </wsa10:Address>
            </wsa10:EndpointReference>
        </wsdl:port>
    </wsdl:service>
</wsdl:definitions>"""


class FaultDefinitionsFilterTestCase(test.TestCase):
    """Tests for the plugin to process the DataCore API WSDL document."""

    @staticmethod
    def _binding_operation_has_fault(document, operation_name):
        for binding in document.getChildren('binding', wsdl.wsdlns):
            for operation in binding.getChildren('operation', wsdl.wsdlns):
                if operation.get('name') == operation_name:
                    fault = operation.getChildren('fault', wsdl.wsdlns)
                    if fault:
                        return True
        return False

    @staticmethod
    def _port_type_operation_has_fault(document, operation_name):
        for port_type in document.getChildren('portType', wsdl.wsdlns):
            for operation in port_type.getChildren('operation', wsdl.wsdlns):
                if operation.get('name') == operation_name:
                    fault = operation.getChildren('fault', wsdl.wsdlns)
                    if fault:
                        return True
        return False

    def _operation_has_fault(self, document, operation_name):
        _binding_has_fault = self._binding_operation_has_fault(
            document, operation_name)
        _port_type_has_fault = self._port_type_operation_has_fault(
            document, operation_name)
        self.assertEqual(_binding_has_fault, _port_type_has_fault)
        return _binding_has_fault

    def test_parsed(self):
        context = mock.Mock()
        sax = parser.Parser()
        wsdl_document = FAKE_WSDL_DOCUMENT
        if isinstance(wsdl_document, str):
            wsdl_document = wsdl_document.encode('utf-8')
        context.document = sax.parse(string=wsdl_document).root()
        self.assertTrue(self._operation_has_fault(context.document,
                                                  'StartExecutive'))
        self.assertTrue(self._operation_has_fault(context.document,
                                                  'StopExecutive'))
        self.assertTrue(self._operation_has_fault(context.document,
                                                  'ExecutiveStarted'))
        self.assertTrue(self._operation_has_fault(context.document,
                                                  'ExecutiveStopped'))
        plugin = api.FaultDefinitionsFilter()
        plugin.parsed(context)
        self.assertTrue(self._operation_has_fault(context.document,
                                                  'StartExecutive'))
        self.assertTrue(self._operation_has_fault(context.document,
                                                  'StopExecutive'))
        self.assertFalse(self._operation_has_fault(context.document,
                                                   'ExecutiveStarted'))
        self.assertFalse(self._operation_has_fault(context.document,
                                                   'ExecutiveStopped'))



#    Copyright 2015 Intel Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import timeutils
from oslo_versionedobjects import fields

from cinder.db.sqlalchemy import models
from cinder import objects


def fake_service_orm(**updates):
    """Create a fake ORM service instance."""
    db_service = fake_db_service(**updates)
    service = models.Service(**db_service)
    return service


def fake_db_service(**updates):
    NOW = timeutils.utcnow().replace(microsecond=0)
    db_service = {
        'created_at': NOW,
        'updated_at': NOW,
        'deleted_at': None,
        'deleted': False,
        'id': 123,
        'uuid': 'ce59413f-4061-425c-9ad0-3479bd102ab2',
        'host': 'fake-host',
        'binary': 'fake-service',
        'topic': 'fake-service-topic',
        'report_count': 1,
        'disabled': False,
        'disabled_reason': None,
        'modified_at': NOW,
    }

    for name, field in objects.Service.fields.items():
        if name in db_service:
            continue
        if field.nullable:
            db_service[name] = None
        elif field.default != fields.UnspecifiedDefault:
            db_service[name] = field.default
        else:
            raise Exception('fake_db_service needs help with %s.' % name)

    if updates:
        db_service.update(updates)

    return db_service


def fake_service_obj(context, **updates):
    return objects.Service._from_db_object(context, objects.Service(),
                                           fake_db_service(**updates))



#    Copyright (c) 2020 Open-E, Inc.
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from cinder import exception
from cinder.i18n import _


class JDSSException(exception.VolumeDriverException):
    """Unknown error"""
    message = _("%(reason)s")


class JDSSRESTException(JDSSException):
    """Unknown communication error"""

    message = _("JDSS REST request %(request)s faild: %(reason)s.")


class JDSSRESTProxyException(JDSSException):
    """Connection with host failed"""

    message = _("JDSS connection with %(host)s failed: %(reason)s.")


class JDSSResourceNotFoundException(JDSSException):
    """Resource does not exist"""

    message = _("JDSS resource %(res)s DNE.")


class JDSSVolumeNotFoundException(JDSSResourceNotFoundException):
    """Volume does not exist"""

    message = _("JDSS volume %(volume)s DNE.")


class JDSSSnapshotNotFoundException(JDSSResourceNotFoundException):
    """Snapshot does not exist"""

    message = _("JDSS snapshot %(snapshot)s DNE.")


class JDSSResourceExistsException(JDSSException):
    """Resource with specified id exists"""

    message = _("JDSS resource with id %(res)s exists.")


class JDSSSnapshotExistsException(JDSSResourceExistsException):
    """Snapshot with the same id exists"""

    message = _("JDSS snapshot %(snapshot)s already exists.")


class JDSSVolumeExistsException(JDSSResourceExistsException):
    """Volume with same id exists"""

    message = _("JDSS volume %(volume)s already exists.")


class JDSSResourceIsBusyException(JDSSException):
    """Resource have dependents"""

    message = _("JDSS resource %(res)s is busy.")


class JDSSSnapshotIsBusyException(JDSSResourceIsBusyException):
    """Snapshot have dependent clones"""

    message = _("JDSS snapshot %(snapshot)s is busy.")


class JDSSOSException(JDSSException):
    """Storage internal system error"""

    message = _("JDSS internal system error %(message)s.")



# Copyright (C) 2017 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Schema for V3 Attachments API.

"""
import copy

from cinder.api.validation import parameter_types


create = {
    'type': 'object',
    'properties': {
        'type': 'object',
        'attachment': {
            'type': 'object',
            'properties': {
                'instance_uuid': parameter_types.uuid,
                'connector': {'type': ['object', 'null']},
                'volume_uuid': parameter_types.uuid,
            },
            'required': ['volume_uuid'],
            'additionalProperties': False,
        },
    },
    'required': ['attachment'],
    'additionalProperties': False,
}

update = {
    'type': 'object',
    'properties': {
        'type': 'object',
        'attachment': {
            'type': 'object',
            'properties': {
                'connector': {'type': 'object', 'minProperties': 1},
            },
            'required': ['connector'],
            'additionalProperties': False,
        },
    },
    'required': ['attachment'],
    'additionalProperties': False,
}

create_v354 = copy.deepcopy(create)
create_v354['properties']['attachment']['properties']['mode'] = (
    {'type': 'string', 'enum': ['rw', 'ro']})



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Stochastic weight handler

This weight handler differs from the default weight
handler by giving every pool a chance to be chosen
where the probability is proportional to each pools'
weight.
"""

import random

from cinder.scheduler import base_weight
from cinder.scheduler import weights as wts


class StochasticHostWeightHandler(base_weight.BaseWeightHandler):
    def __init__(self, namespace):
        super(StochasticHostWeightHandler, self).__init__(wts.BaseHostWeigher,
                                                          namespace)

    def get_weighed_objects(self, weigher_classes, obj_list,
                            weighing_properties):
        # The normalization performed in the superclass is nonlinear, which
        # messes up the probabilities, so override it. The probabilistic
        # approach we use here is self-normalizing.
        # Also, the sorting done by the parent implementation is harmless but
        # useless for us.

        # Compute the object weights as the parent would but without sorting
        # or normalization.
        weighed_objs = [wts.WeighedHost(obj, 0.0) for obj in obj_list]
        for weigher_cls in weigher_classes:
            weigher = weigher_cls()
            weights = weigher.weigh_objects(weighed_objs, weighing_properties)
            for i, weight in enumerate(weights):
                obj = weighed_objs[i]
                obj.weight += weigher.weight_multiplier() * weight

        # Avoid processing empty lists
        if not weighed_objs:
            return []

        # First compute the total weight of all the objects and the upper
        # bound for each object to "win" the lottery.
        total_weight = 0.0
        table = []
        for weighed_obj in weighed_objs:
            total_weight += weighed_obj.weight
            max_value = total_weight
            table.append((max_value, weighed_obj))

        # Now draw a random value with the computed range
        winning_value = random.random() * total_weight

        # Scan the table to find the first object with a maximum higher than
        # the random number. Save the index of the winner.
        winning_index = 0
        for (i, (max_value, weighed_obj)) in enumerate(table):
            if max_value > winning_value:
                # Return a single element array with the winner.
                winning_index = i
                break

        # It's theoretically possible for the above loop to terminate with no
        # winner. This happens when winning_value >= total_weight, which
        # could only occur with very large numbers and floating point
        # rounding. In those cases the actual winner should have been the
        # last element, so return it.
        return weighed_objs[winning_index:] + weighed_objs[0:winning_index]



# Copyright (C) 2012 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from http import client as http_client
import json
import os
import socket
import zlib

from swiftclient import client as swift


class FakeSwiftClient(object):
    """Logs calls instead of executing."""
    def __init__(self, *args, **kwargs):
        pass

    @classmethod
    def Connection(cls, *args, **kargs):
        return FakeSwiftConnection()


class FakeSwiftConnection(object):
    """Logging calls instead of executing."""
    def __init__(self, *args, **kwargs):
        pass

    def head_container(self, container, headers=None):
        if container in ['missing_container',
                         'missing_container_socket_error_on_put']:
            raise swift.ClientException('fake exception',
                                        http_status=http_client.NOT_FOUND)
        elif container == 'unauthorized_container':
            raise swift.ClientException('fake exception',
                                        http_status=http_client.UNAUTHORIZED)
        elif container == 'socket_error_on_head':
            raise socket.error(111, 'ECONNREFUSED')
        pass

    def put_container(self, container, headers=None):
        if container == 'missing_container_socket_error_on_put':
            raise socket.error(111, 'ECONNREFUSED')

    def get_container(self, container, headers=None, **kwargs):
        fake_header = None
        fake_body = [{'name': 'backup_001'},
                     {'name': 'backup_002'},
                     {'name': 'backup_003'}]
        return fake_header, fake_body

    def head_object(self, container, name, headers=None):
        return {'etag': 'fake-md5-sum'}

    def get_object(self, container, name, headers=None):
        if container == 'socket_error_on_get':
            raise socket.error(111, 'ECONNREFUSED')
        if 'metadata' in name:
            fake_object_header = None
            metadata = {}
            if container == 'unsupported_version':
                metadata['version'] = '9.9.9'
            else:
                metadata['version'] = '1.0.0'
            metadata['backup_id'] = 123
            metadata['volume_id'] = 123
            metadata['backup_name'] = 'fake backup'
            metadata['backup_description'] = 'fake backup description'
            metadata['created_at'] = '2013-02-19 11:20:54,805'
            metadata['objects'] = [{
                'backup_001': {'compression': 'zlib', 'length': 10,
                               'offset': 0},
                'backup_002': {'compression': 'zlib', 'length': 10,
                               'offset': 10},
                'backup_003': {'compression': 'zlib', 'length': 10,
                               'offset': 20}
            }]
            metadata_json = json.dumps(metadata, sort_keys=True, indent=2)
            metadata_json = metadata_json.encode('utf-8')
            fake_object_body = metadata_json
            return (fake_object_header, fake_object_body)

        fake_header = None
        fake_object_body = os.urandom(1024 * 1024)
        return (fake_header, zlib.compress(fake_object_body))

    def put_object(self, container, name, reader, content_length=None,
                   etag=None, chunk_size=None, content_type=None,
                   headers=None, query_string=None):
        if container == 'socket_error_on_put':
            raise socket.error(111, 'ECONNREFUSED')
        return 'fake-md5-sum'

    def delete_object(self, container, name, headers=None):
        if container == 'socket_error_on_delete':
            raise socket.error(111, 'ECONNREFUSED')
        pass



# -*- coding: utf-8 -*-
# Copyright 2014, Adrien Vergé <adrien.verge@numergy.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_log import log as logging
from oslo_utils import uuidutils

from cinder.compute import nova
from cinder import exception
from cinder.i18n import _
from cinder.scheduler import filters
from cinder.volume import volume_utils


LOG = logging.getLogger(__name__)

HINT_KEYWORD = 'local_to_instance'
INSTANCE_HOST_PROP = 'OS-EXT-SRV-ATTR:host'
REQUESTS_TIMEOUT = 5


class InstanceLocalityFilter(filters.BaseBackendFilter):
    """Schedule volume on the same host as a given instance.

    This filter enables selection of a storage back-end located on the host
    where the instance's hypervisor is running. This provides data locality:
    the instance and the volume are located on the same physical machine.

    In order to work:

    - The Extended Server Attributes extension needs to be active in Nova (this
      is by default), so that the 'OS-EXT-SRV-ATTR:host' property is returned
      when requesting instance info.
    - Either an account with privileged rights for Nova must be configured in
      Cinder configuration (configure a keystone authentication plugin in the
      [nova] section), or the user making the call needs to have sufficient
      rights (see 'extended_server_attributes' in Nova policy).

    """

    def __init__(self):
        # Cache Nova API answers directly into the Filter object.
        # Since a BaseBackendFilter instance lives only during the volume's
        # scheduling, the cache is re-created for every new volume creation.
        self._cache = {}
        super(InstanceLocalityFilter, self).__init__()

    def backend_passes(self, backend_state, filter_properties):
        context = filter_properties['context']
        backend = volume_utils.extract_host(backend_state.backend_id, 'host')

        scheduler_hints = filter_properties.get('scheduler_hints') or {}
        instance_uuid = scheduler_hints.get(HINT_KEYWORD, None)

        # Without 'local_to_instance' hint
        if not instance_uuid:
            return True

        if not uuidutils.is_uuid_like(instance_uuid):
            raise exception.InvalidUUID(uuid=instance_uuid)

        # TODO(adrienverge): Currently it is not recommended to allow instance
        # migrations for hypervisors where this hint will be used. In case of
        # instance migration, a previously locally-created volume will not be
        # automatically migrated. Also in case of instance migration during the
        # volume's scheduling, the result is unpredictable. A future
        # enhancement would be to subscribe to Nova migration events (e.g. via
        # Ceilometer).

        # First, lookup for already-known information in local cache
        if instance_uuid in self._cache:
            return self._cache[instance_uuid] == backend

        server = nova.API().get_server(context, instance_uuid,
                                       privileged_user=True,
                                       timeout=REQUESTS_TIMEOUT)

        if not hasattr(server, INSTANCE_HOST_PROP):
            LOG.warning('Hint "%s" dropped because Nova did not return '
                        'enough information. Either Nova policy needs to '
                        'be changed or a privileged account for Nova '
                        'should be specified in conf.', HINT_KEYWORD)
            raise exception.CinderException(_('Hint "%s" not supported.') %
                                            HINT_KEYWORD)

        self._cache[instance_uuid] = getattr(server, INSTANCE_HOST_PROP)

        # Match if given instance is hosted on backend
        return self._cache[instance_uuid] == backend



# Copyright (c) 2022 DataCore Software Corp. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Exception definitions."""

from cinder import exception
from cinder.i18n import _


class DataCoreException(exception.VolumeBackendAPIException):
    """Base DataCore Exception."""

    message = _('DataCore exception.')


class DataCoreConnectionException(DataCoreException):
    """Thrown when there are connection problems during a DataCore API call."""

    message = _('Failed to connect to DataCore Server Group: %(reason)s.')


class DataCoreFaultException(DataCoreException):
    """Thrown when there are faults during a DataCore API call."""

    message = _('DataCore Server Group reported an error: %(reason)s.')



# Copyright (c) 2016 Huawei Technologies Co., Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Sets Huawei private configuration into Configuration object.

For conveniently get private configuration. We parse Huawei config file
and set every property into Configuration object as an attribute.
"""

import base64
import os
import re

from lxml import etree as ET
from oslo_log import log as logging
import six

from cinder import exception
from cinder.i18n import _
from cinder.volume.drivers.huawei import constants

LOG = logging.getLogger(__name__)


class HuaweiConf(object):
    def __init__(self, conf):
        self.conf = conf
        self.last_modify_time = None

    def update_config_value(self):
        file_time = os.stat(self.conf.cinder_huawei_conf_file).st_mtime
        if self.last_modify_time == file_time:
            return

        self.last_modify_time = file_time
        tree = ET.parse(self.conf.cinder_huawei_conf_file)
        xml_root = tree.getroot()
        self._encode_authentication(tree, xml_root)

        attr_funcs = (
            self._san_address,
            self._san_user,
            self._san_password,
            self._san_vstore,
            self._san_product,
            self._ssl_cert_path,
            self._ssl_cert_verify,
            self._iscsi_info,
            self._fc_info,
            self._hypermetro_devices,
            self._replication_devices,
            self._lun_type,
            self._lun_ready_wait_interval,
            self._lun_copy_wait_interval,
            self._lun_copy_speed,
            self._lun_timeout,
            self._lun_write_type,
            self._lun_prefetch,
            self._lun_policy,
            self._lun_read_cache_policy,
            self._lun_write_cache_policy,
            self._storage_pools,
            self._get_local_minimum_fc_initiator,
        )

        for f in attr_funcs:
            f(xml_root)

    def _encode_authentication(self, tree, xml_root):
        name_node = xml_root.find('Storage/UserName')
        pwd_node = xml_root.find('Storage/UserPassword')
        vstore_node = xml_root.find('Storage/vStoreName')

        need_encode = False
        if name_node is not None and not name_node.text.startswith('!$$$'):
            encoded = base64.b64encode(six.b(name_node.text)).decode()
            name_node.text = '!$$$' + encoded
            need_encode = True

        if pwd_node is not None and not pwd_node.text.startswith('!$$$'):
            encoded = base64.b64encode(six.b(pwd_node.text)).decode()
            pwd_node.text = '!$$$' + encoded
            need_encode = True

        if vstore_node is not None and not vstore_node.text.startswith('!$$$'):
            encoded = base64.b64encode(six.b(vstore_node.text)).decode()
            vstore_node.text = '!$$$' + encoded
            need_encode = True

        if need_encode:
            tree.write(self.conf.cinder_huawei_conf_file, encoding='UTF-8')

    def _san_address(self, xml_root):
        text = xml_root.findtext('Storage/RestURL')
        if not text:
            msg = _("RestURL is not configured.")
            LOG.error(msg)
            raise exception.InvalidInput(reason=msg)

        addrs = list(set([x.strip() for x in text.split(';') if x.strip()]))
        setattr(self.conf, 'san_address', addrs)

    def _san_user(self, xml_root):
        text = xml_root.findtext('Storage/UserName')
        if not text:
            msg = _("UserName is not configured.")
            LOG.error(msg)
            raise exception.InvalidInput(reason=msg)

        user = base64.b64decode(six.b(text[4:])).decode()
        setattr(self.conf, 'san_user', user)

    def _san_password(self, xml_root):
        text = xml_root.findtext('Storage/UserPassword')
        if not text:
            msg = _("UserPassword is not configured.")
            LOG.error(msg)
            raise exception.InvalidInput(reason=msg)

        pwd = base64.b64decode(six.b(text[4:])).decode()
        setattr(self.conf, 'san_password', pwd)

    def _san_vstore(self, xml_root):
        vstore = None
        text = xml_root.findtext('Storage/vStoreName')
        if text:
            vstore = base64.b64decode(six.b(text[4:])).decode()
        setattr(self.conf, 'vstore_name', vstore)

    def _ssl_cert_path(self, xml_root):
        text = xml_root.findtext('Storage/SSLCertPath')
        setattr(self.conf, 'ssl_cert_path', text)

    def _ssl_cert_verify(self, xml_root):
        value = False
        text = xml_root.findtext('Storage/SSLCertVerify')
        if text:
            if text.lower() in ('true', 'false'):
                value = text.lower() == 'true'
            else:
                msg = _("SSLCertVerify configured error.")
                LOG.error(msg)
                raise exception.InvalidInput(reason=msg)

        setattr(self.conf, 'ssl_cert_verify', value)

    def _set_extra_constants_by_product(self, product):
        extra_constants = {}
        if product == 'Dorado':
            extra_constants['QOS_SPEC_KEYS'] = (
                'maxIOPS', 'maxBandWidth', 'IOType')
            extra_constants['QOS_IOTYPES'] = ('2',)
            extra_constants['SUPPORT_LUN_TYPES'] = ('Thin',)
            extra_constants['DEFAULT_LUN_TYPE'] = 'Thin'
        else:
            extra_constants['QOS_SPEC_KEYS'] = (
                'maxIOPS', 'minIOPS', 'minBandWidth',
                'maxBandWidth', 'latency', 'IOType')
            extra_constants['QOS_IOTYPES'] = ('0', '1', '2')
            extra_constants['SUPPORT_LUN_TYPES'] = ('Thick', 'Thin')
            extra_constants['DEFAULT_LUN_TYPE'] = 'Thick'

        for k in extra_constants:
            setattr(constants, k, extra_constants[k])

    def _san_product(self, xml_root):
        text = xml_root.findtext('Storage/Product')
        if not text:
            msg = _("SAN product is not configured.")
            LOG.error(msg)
            raise exception.InvalidInput(reason=msg)

        product = text.strip()
        if product not in constants.VALID_PRODUCT:
            msg = _("Invalid SAN product %(text)s, SAN product must be "
                    "in %(valid)s.") % {'text': product,
                                        'valid': constants.VALID_PRODUCT}
            LOG.error(msg)
            raise exception.InvalidInput(reason=msg)

        self._set_extra_constants_by_product(product)
        setattr(self.conf, 'san_product', product)

    def _lun_type(self, xml_root):
        lun_type = constants.DEFAULT_LUN_TYPE
        text = xml_root.findtext('LUN/LUNType')
        if text:
            lun_type = text.strip()
            if lun_type not in constants.LUN_TYPE_MAP:
                msg = _("Invalid lun type %s is configured.") % lun_type
                LOG.error(msg)
                raise exception.InvalidInput(reason=msg)

            if lun_type not in constants.SUPPORT_LUN_TYPES:
                msg = _("%(array)s array requires %(valid)s lun type, "
                        "but %(conf)s is specified."
                        ) % {'array': self.conf.san_product,
                             'valid': constants.SUPPORT_LUN_TYPES,
                             'conf': lun_type}
                LOG.error(msg)
                raise exception.InvalidInput(reason=msg)

        setattr(self.conf, 'lun_type', constants.LUN_TYPE_MAP[lun_type])

    def _lun_ready_wait_interval(self, xml_root):
        text = xml_root.findtext('LUN/LUNReadyWaitInterval')
        interval = text.strip() if text else constants.DEFAULT_WAIT_INTERVAL
        setattr(self.conf, 'lun_ready_wait_interval', int(interval))

    def _lun_copy_wait_interval(self, xml_root):
        text = xml_root.findtext('LUN/LUNcopyWaitInterval')
        interval = text.strip() if text else constants.DEFAULT_WAIT_INTERVAL
        setattr(self.conf, 'lun_copy_wait_interval', int(interval))

    def _lun_timeout(self, xml_root):
        text = xml_root.findtext('LUN/Timeout')
        interval = text.strip() if text else constants.DEFAULT_WAIT_TIMEOUT
        setattr(self.conf, 'lun_timeout', int(interval))

    def _lun_write_type(self, xml_root):
        text = xml_root.findtext('LUN/WriteType')
        if text and text.strip():
            setattr(self.conf, 'write_type', text.strip())

    def _lun_prefetch(self, xml_root):
        node = xml_root.find('LUN/Prefetch')
        if node is not None:
            if 'Type' in node.attrib:
                prefetch_type = node.attrib['Type'].strip()
                setattr(self.conf, 'prefetch_type', prefetch_type)

            if 'Value' in node.attrib:
                prefetch_value = node.attrib['Value'].strip()
                setattr(self.conf, 'prefetch_value', prefetch_value)

    def _lun_policy(self, xml_root):
        setattr(self.conf, 'lun_policy', '0')

    def _lun_read_cache_policy(self, xml_root):
        setattr(self.conf, 'lun_read_cache_policy', '2')

    def _lun_write_cache_policy(self, xml_root):
        setattr(self.conf, 'lun_write_cache_policy', '5')

    def _storage_pools(self, xml_root):
        text = xml_root.findtext('LUN/StoragePool')
        if not text:
            msg = _('Storage pool is not configured.')
            LOG.error(msg)
            raise exception.InvalidInput(reason=msg)

        pools = set(x.strip() for x in text.split(';') if x.strip())
        if not pools:
            msg = _('No valid storage pool configured.')
            LOG.error(msg)
            raise exception.InvalidInput(msg)

        setattr(self.conf, 'storage_pools', list(pools))

    def _iscsi_info(self, xml_root):
        iscsi_info = {
            'default_target_ips': [],
            'CHAPinfo': xml_root.findtext('iSCSI/CHAPinfo'),
            'ALUA': xml_root.findtext('iSCSI/ALUA'),
            'FAILOVERMODE': xml_root.findtext('iSCSI/FAILOVERMODE'),
            'SPECIALMODETYPE': xml_root.findtext('iSCSI/SPECIALMODETYPE'),
            'PATHTYPE': xml_root.findtext('iSCSI/PATHTYPE'),
        }

        text = xml_root.findtext('iSCSI/DefaultTargetIP')
        if text:
            iscsi_info['default_target_ips'] = [
                ip.strip() for ip in text.split(';') if ip.strip()]

        initiators = {}
        nodes = xml_root.findall('iSCSI/Initiator')
        for node in nodes or []:
            if 'Name' not in node.attrib:
                msg = _('Name must be specified for initiator.')
                LOG.error(msg)
                raise exception.InvalidInput(msg)

            initiators[node.attrib['Name']] = node.attrib

        iscsi_info['initiators'] = initiators
        setattr(self.conf, 'iscsi_info', iscsi_info)

    def _fc_info(self, xml_root):
        fc_info = {
            'ALUA': xml_root.findtext('FC/ALUA'),
            'FAILOVERMODE': xml_root.findtext('FC/FAILOVERMODE'),
            'SPECIALMODETYPE': xml_root.findtext('FC/SPECIALMODETYPE'),
            'PATHTYPE': xml_root.findtext('FC/PATHTYPE'),
        }

        initiators = {}
        nodes = xml_root.findall('FC/Initiator')
        for node in nodes or []:
            if 'Name' not in node.attrib:
                msg = _('Name must be specified for initiator.')
                LOG.error(msg)
                raise exception.InvalidInput(msg)

            initiators[node.attrib['Name']] = node.attrib

        fc_info['initiators'] = initiators
        setattr(self.conf, 'fc_info', fc_info)

    def _parse_remote_initiator_info(self, dev, ini_type):
        ini_info = {'default_target_ips': []}

        if dev.get('iscsi_default_target_ip'):
            ini_info['default_target_ips'] = dev[
                'iscsi_default_target_ip'].split(';')

        initiators = {}
        if ini_type in dev:
            # Analyze initiators configure text, convert to:
            # [{'Name':'xxx'}, {'Name':'xxx','CHAPinfo':'mm-usr#mm-pwd'}]
            ini_list = re.split(r'\s', dev[ini_type])

            def _convert_one_iscsi_info(ini_text):
                # get initiator configure attr list
                attr_list = re.split('[{;}]', ini_text)

                # get initiator configures
                ini = {}
                for attr in attr_list:
                    if not attr:
                        continue

                    pair = attr.split(':', 1)
                    if pair[0] == 'CHAPinfo':
                        value = pair[1].replace('#', ';', 1)
                    else:
                        value = pair[1]
                    ini[pair[0]] = value

                if 'Name' not in ini:
                    msg = _('Name must be specified for initiator.')
                    LOG.error(msg)
                    raise exception.InvalidInput(msg)

                return ini

            for text in ini_list:
                ini = _convert_one_iscsi_info(text)
                initiators[ini['Name']] = ini

        ini_info['initiators'] = initiators
        return ini_info

    def _hypermetro_devices(self, xml_root):
        dev = self.conf.safe_get('hypermetro_device')
        config = {}

        if dev:
            config = {
                'san_address': dev['san_address'].split(';'),
                'san_user': dev['san_user'],
                'san_password': dev['san_password'],
                'vstore_name': dev.get('vstore_name'),
                'metro_domain': dev['metro_domain'],
                'storage_pools': dev['storage_pool'].split(';')[:1],
                'iscsi_info': self._parse_remote_initiator_info(
                    dev, 'iscsi_info'),
                'fc_info': self._parse_remote_initiator_info(
                    dev, 'fc_info'),
            }

        setattr(self.conf, 'hypermetro', config)

    def _replication_devices(self, xml_root):
        replication_devs = self.conf.safe_get('replication_device')
        config = {}

        if replication_devs:
            dev = replication_devs[0]
            config = {
                'backend_id': dev['backend_id'],
                'san_address': dev['san_address'].split(';'),
                'san_user': dev['san_user'],
                'san_password': dev['san_password'],
                'vstore_name': dev.get('vstore_name'),
                'storage_pools': dev['storage_pool'].split(';')[:1],
                'iscsi_info': self._parse_remote_initiator_info(
                    dev, 'iscsi_info'),
                'fc_info': self._parse_remote_initiator_info(
                    dev, 'fc_info'),
            }

        setattr(self.conf, 'replication', config)

    def _lun_copy_speed(self, xml_root):
        text = xml_root.findtext('LUN/LUNCopySpeed')
        if text and text.strip() not in constants.LUN_COPY_SPEED_TYPES:
            msg = (_("Invalid LUNCopySpeed '%(text)s', LUNCopySpeed must "
                     "be between %(low)s and %(high)s.")
                   % {"text": text, "low": constants.LUN_COPY_SPEED_LOW,
                      "high": constants.LUN_COPY_SPEED_HIGHEST})
            LOG.error(msg)
            raise exception.InvalidInput(reason=msg)

        if not text:
            speed = constants.LUN_COPY_SPEED_MEDIUM
        else:
            speed = text.strip()
        setattr(self.conf, 'lun_copy_speed', int(speed))

    def _get_local_minimum_fc_initiator(self, xml_root):
        text = xml_root.findtext('FC/MinOnlineFCInitiator')
        minimum_fc_initiator = constants.DEFAULT_MINIMUM_FC_INITIATOR_ONLINE

        if not text:
            LOG.info("MinOnlineFCInitiator not set, using default.")
            setattr(self.conf, 'min_fc_ini_online', minimum_fc_initiator)
            return

        text = text.strip()
        if not text.isdigit():
            msg = (_("Invalid FC MinOnlineFCInitiator '%s', "
                     "MinOnlineFCInitiator must be a digit.") % text)
            LOG.error(msg)
            raise exception.InvalidInput(reason=msg)

        minimum_fc_initiator = int(text)
        setattr(self.conf, 'min_fc_ini_online', minimum_fc_initiator)



# Copyright 2016 IBM Corp.
# Copyright 2017 Rackspace Australia
# Copyright 2018 Michael Still and Aptira
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Fixtures for Cinder tests."""

import logging as std_logging
import os
import warnings

import fixtures
from oslo_config import cfg
from oslo_policy import policy as oslo_policy
from oslo_privsep import daemon as privsep_daemon
from sqlalchemy import exc as sqla_exc

import cinder.policy

CONF = cfg.CONF
_TRUE_VALUES = ('True', 'true', '1', 'yes')


class NullHandler(std_logging.Handler):
    """custom default NullHandler to attempt to format the record.

    Used in conjunction with
    log_fixture.get_logging_handle_error_fixture to detect formatting errors in
    debug level logs without saving the logs.
    """
    def handle(self, record):
        self.format(record)

    def emit(self, record):
        pass

    def createLock(self):
        self.lock = None


class StandardLogging(fixtures.Fixture):
    """Setup Logging redirection for tests.

    There are a number of things we want to handle with logging in tests:

    * Redirect the logging to somewhere that we can test or dump it later.

    * Ensure that as many DEBUG messages as possible are actually
       executed, to ensure they are actually syntactically valid (they
       often have not been).

    * Ensure that we create useful output for tests that doesn't
      overwhelm the testing system (which means we can't capture the
      100 MB of debug logging on every run).

    To do this we create a logger fixture at the root level, which
    defaults to INFO and create a Null Logger at DEBUG which lets
    us execute log messages at DEBUG but not keep the output.

    To support local debugging OS_DEBUG=True can be set in the
    environment, which will print out the full debug logging.

    There are also a set of overrides for particularly verbose
    modules to be even less than INFO.

    """

    def setUp(self):
        super(StandardLogging, self).setUp()

        # set root logger to debug
        root = std_logging.getLogger()
        root.setLevel(std_logging.INFO)

        # supports collecting debug level for local runs
        if os.environ.get('OS_DEBUG') in _TRUE_VALUES:
            level = std_logging.DEBUG
        else:
            level = std_logging.INFO

        # Collect logs
        fs = '%(asctime)s %(levelname)s [%(name)s] %(message)s'
        self.logger = self.useFixture(
            fixtures.FakeLogger(format=fs, level=None))
        # TODO(sdague): why can't we send level through the fake
        # logger? Tests prove that it breaks, but it's worth getting
        # to the bottom of.
        root.handlers[0].setLevel(level)

        if level > std_logging.DEBUG:
            # Just attempt to format debug level logs, but don't save them
            handler = NullHandler()
            self.useFixture(fixtures.LogHandler(handler, nuke_handlers=False))
            handler.setLevel(std_logging.DEBUG)

            # Don't log every single DB migration step
            std_logging.getLogger(
                'migrate.versioning.api').setLevel(std_logging.WARNING)

        # At times we end up calling back into main() functions in
        # testing. This has the possibility of calling logging.setup
        # again, which completely unwinds the logging capture we've
        # created here. Once we've setup the logging in the way we want,
        # disable the ability for the test to change this.
        def fake_logging_setup(*args):
            pass

        self.useFixture(
            fixtures.MonkeyPatch('oslo_log.log.setup', fake_logging_setup))


class WarningsFixture(fixtures.Fixture):
    """Filters out warnings during test runs."""

    def setUp(self):
        super().setUp()

        self._original_warning_filters = warnings.filters[:]

        # NOTE(sdague): Make deprecation warnings only happen once. Otherwise
        # this gets kind of crazy given the way that upstream python libs use
        # this.
        warnings.simplefilter('once', DeprecationWarning)

        # NOTE(sdague): this remains an unresolved item around the way
        # forward on is_admin, the deprecation is definitely really premature.
        warnings.filterwarnings(
            'ignore',
            message='Policy enforcement is depending on the value of is_admin.'
                    ' This key is deprecated. Please update your policy '
                    'file to use the standard policy values.')

        # We can't do anything about this outside of cinder
        warnings.filterwarnings(
            'ignore',
            message='distutils Version classes are deprecated. .*',
            category=DeprecationWarning,
        )

        warnings.filterwarnings(
            'ignore',
            message='the imp module is deprecated in favour of importlib',
            category=DeprecationWarning,
        )

        warnings.filterwarnings(
            'ignore',
            message='invalid escape sequence',
            category=DeprecationWarning,
        )
        warnings.filterwarnings(
            'error',
            message='invalid escape sequence',
            category=DeprecationWarning,
            module='cinder',
        )

        warnings.filterwarnings(
            'ignore',
            message='Policy ".*":".*" was deprecated in ',
            module='oslo_policy',
            category=UserWarning,
        )

        # Enable deprecation warnings for cinder itself to capture upcoming
        # SQLAlchemy changes

        warnings.filterwarnings(
            'ignore',
            category=sqla_exc.SADeprecationWarning,
        )

        warnings.filterwarnings(
            'error',
            module='cinder',
            category=sqla_exc.SADeprecationWarning,
        )

        # ...but filter everything out until we get around to fixing them
        # TODO: Fix all of these

        warnings.filterwarnings(
            'ignore',
            module='cinder',
            message='The current statement is being autocommitted using ',
            category=sqla_exc.SADeprecationWarning,
        )

        warnings.filterwarnings(
            'ignore',
            module='cinder',
            message='The Session.begin.subtransactions flag is deprecated ',
            category=sqla_exc.SADeprecationWarning,
        )

        self.addCleanup(self._reset_warning_filters)

    def _reset_warning_filters(self):
        warnings.filters[:] = self._original_warning_filters


class UnHelperfulClientChannel(privsep_daemon._ClientChannel):
    def __init__(self, context):
        raise Exception('You have attempted to start a privsep helper. '
                        'This is not allowed in the gate, and '
                        'indicates a failure to have mocked your tests.')


class PrivsepNoHelperFixture(fixtures.Fixture):
    """A fixture to catch failures to mock privsep's rootwrap helper.

    If you fail to mock away a privsep'd method in a unit test, then
    you may well end up accidentally running the privsep rootwrap
    helper. This will fail in the gate, but it fails in a way which
    doesn't identify which test is missing a mock. Instead, we
    raise an exception so that you at least know where you've missed
    something.
    """

    def setUp(self):
        super(PrivsepNoHelperFixture, self).setUp()

        self.useFixture(fixtures.MonkeyPatch(
            'oslo_privsep.daemon.RootwrapClientChannel',
            UnHelperfulClientChannel))


class PolicyFixture(fixtures.Fixture):
    """Load the live policy for tests.

    A base policy fixture that starts with the assumption that you'd
    like to load and enforce the shipped default policy in tests.

    """
    def setUp(self):
        super().setUp()
        cinder.policy.reset()
        # Suppress deprecation warnings for unit tests.
        cinder.policy.init(suppress_deprecation_warnings=True)
        self.addCleanup(cinder.policy.reset)

    def set_rules(self, rules, overwrite=True):
        policy = cinder.policy._ENFORCER
        policy.set_rules(oslo_policy.Rules.from_dict(rules),
                         overwrite=overwrite)



# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Cinder base exception handling.

Includes decorator for re-raising Cinder-type exceptions.

SHOULD include dedicated exception logging.

"""

from typing import Optional, Union  # noqa: H301

from oslo_log import log as logging
from oslo_versionedobjects import exception as obj_exc
import webob.exc
from webob.util import status_generic_reasons
from webob.util import status_reasons

from cinder.i18n import _


LOG = logging.getLogger(__name__)


class ConvertedException(webob.exc.WSGIHTTPException):
    def __init__(self, code: int = 500, title: str = "",
                 explanation: str = ""):
        self.code = code
        # There is a strict rule about constructing status line for HTTP:
        # '...Status-Line, consisting of the protocol version followed by a
        # numeric status code and its associated textual phrase, with each
        # element separated by SP characters'
        # (http://www.faqs.org/rfcs/rfc2616.html)
        # 'code' and 'title' can not be empty because they correspond
        # to numeric status code and its associated text
        if title:
            self.title = title
        else:
            try:
                self.title = status_reasons[self.code]
            except KeyError:
                generic_code = self.code // 100
                self.title = status_generic_reasons[generic_code]
        self.explanation = explanation
        super(ConvertedException, self).__init__()


class CinderException(Exception):
    """Base Cinder Exception

    To correctly use this class, inherit from it and define
    a 'message' property. That message will get printf'd
    with the keyword arguments provided to the constructor.

    """
    message = _("An unknown exception occurred.")
    code = 500
    headers: dict = {}
    safe = False

    def __init__(self, message: Optional[Union[str, tuple]] = None, **kwargs):
        self.kwargs = kwargs
        self.kwargs['message'] = message

        if 'code' not in self.kwargs:
            try:
                self.kwargs['code'] = self.code
            except AttributeError:
                pass

        for k, v in self.kwargs.items():
            if isinstance(v, Exception):
                # NOTE(tommylikehu): If this is a cinder exception it will
                # return the msg object, so we won't be preventing
                # translations.
                self.kwargs[k] = str(v)

        if self._should_format():
            try:
                message = self.message % kwargs

            except Exception:
                # NOTE(melwitt): This is done in a separate method so it can be
                # monkey-patched during testing to make it a hard failure.
                self._log_exception()
                message = self.message
        elif isinstance(message, Exception):
            # NOTE(tommylikehu): If this is a cinder exception it will
            # return the msg object, so we won't be preventing
            # translations.
            message = str(message)

        # NOTE(luisg): We put the actual message in 'msg' so that we can access
        # it, because if we try to access the message via 'message' it will be
        # overshadowed by the class' message attribute
        self.msg = message
        super(CinderException, self).__init__(message)
        # Oslo.messaging use the argument 'message' to rebuild exception
        # directly at the rpc client side, therefore we should not use it
        # in our keyword arguments, otherwise, the rebuild process will fail
        # with duplicate keyword exception.
        self.kwargs.pop('message', None)

    def _log_exception(self) -> None:
        # kwargs doesn't match a variable in the message
        # log the issue and the kwargs
        LOG.exception('Exception in string format operation:')
        for name, value in self.kwargs.items():
            LOG.error("%(name)s: %(value)s",
                      {'name': name, 'value': value})

    def _should_format(self) -> bool:
        return self.kwargs['message'] is None or '%(message)' in self.message


class VolumeBackendAPIException(CinderException):
    message = _("Bad or unexpected response from the storage volume "
                "backend API: %(data)s")


class VolumeDriverException(CinderException):
    message = _("Volume driver reported an error: %(message)s")


class BackupDriverException(CinderException):
    message = _("Backup driver reported an error: %(reason)s")


class BackupRestoreCancel(CinderException):
    message = _("Canceled backup %(back_id)s restore on volume %(vol_id)s")


class GlanceConnectionFailed(CinderException):
    message = _("Connection to glance failed: %(reason)s")


class ProgrammingError(CinderException):
    message = _('Programming error in Cinder: %(reason)s')


class NotAuthorized(CinderException):
    message = _("Not authorized.")
    code = 403


class AdminRequired(NotAuthorized):
    message = _("User does not have admin privileges")


class PolicyNotAuthorized(NotAuthorized):
    message = _("Policy doesn't allow %(action)s to be performed.")


class ImageNotAuthorized(CinderException):
    message = _("Not authorized for image %(image_id)s.")


class DriverNotInitialized(CinderException):
    message = _("Volume driver not ready.")


class Invalid(CinderException):
    message = _("Unacceptable parameters.")
    code = 400


class InvalidSnapshot(Invalid):
    message = _("Invalid snapshot: %(reason)s")


class InvalidVolumeAttachMode(Invalid):
    message = _("Invalid attaching mode '%(mode)s' for "
                "volume %(volume_id)s.")


class VolumeAttached(Invalid):
    message = _("Volume %(volume_id)s is still attached, detach volume first.")


class InvalidResults(Invalid):
    message = _("The results are invalid.")


class InvalidInput(Invalid):
    message = _("Invalid input received: %(reason)s")


class InvalidAvailabilityZone(Invalid):
    message = _("Availability zone '%(az)s' is invalid.")


class InvalidTypeAvailabilityZones(Invalid):
    message = _("Volume type is only supported in these availability zones: "
                "%(az)s")


class InvalidVolumeType(Invalid):
    message = _("Invalid volume type: %(reason)s")


class InvalidGroupType(Invalid):
    message = _("Invalid group type: %(reason)s")


class InvalidVolume(Invalid):
    message = _("Invalid volume: %(reason)s")


class InvalidContentType(Invalid):
    message = _("Invalid content type %(content_type)s.")


class InvalidHost(Invalid):
    message = _("Invalid host: %(reason)s")


# Cannot be templated as the error syntax varies.
# msg needs to be constructed when raised.
class InvalidParameterValue(Invalid):
    message = "%(err)s"


class InvalidAuthKey(Invalid):
    message = _("Invalid auth key: %(reason)s")


class InvalidConfigurationValue(Invalid):
    message = _('Value "%(value)s" is not valid for '
                'configuration option "%(option)s"')


class ServiceUnavailable(Invalid):
    message = _("Service is unavailable at this time.")


class UnavailableDuringUpgrade(Invalid):
    message = _('Cannot perform %(action)s during system upgrade.')


class ImageUnacceptable(Invalid):
    message = _("Image %(image_id)s is unacceptable: %(reason)s")


class ImageTooBig(Invalid):
    message = _("Image %(image_id)s size exceeded available "
                "disk space: %(reason)s")


class DeviceUnavailable(Invalid):
    message = _("The device in the path %(path)s is unavailable: %(reason)s")


class SnapshotUnavailable(VolumeBackendAPIException):
    message = _("The snapshot is unavailable: %(data)s")


class InvalidUUID(Invalid):
    message = _("Expected a UUID but received %(uuid)s.")


class InvalidAPIVersionString(Invalid):
    message = _("API Version String %(version)s is of invalid format. Must "
                "be of format MajorNum.MinorNum.")


class VersionNotFoundForAPIMethod(Invalid):
    message = _("API version %(version)s is not supported on this method.")


class InvalidGlobalAPIVersion(Invalid):
    message = _("Version %(req_ver)s is not supported by the API. Minimum "
                "is %(min_ver)s and maximum is %(max_ver)s.")


class ValidationError(Invalid):
    message = "%(detail)s"


class APIException(CinderException):
    message = _("Error while requesting %(service)s API.")

    def __init__(self, message=None, **kwargs):
        if 'service' not in kwargs:
            kwargs['service'] = 'unknown'
        super(APIException, self).__init__(message, **kwargs)


class APITimeout(APIException):
    message = _("Timeout while requesting %(service)s API.")


class RPCTimeout(CinderException):
    message = _("Timeout while requesting capabilities from backend "
                "%(service)s.")
    code = 502


class Duplicate(CinderException):
    pass


class NotFound(CinderException):
    message = _("Resource could not be found.")
    code = 404
    safe = True


class GlanceStoreNotFound(NotFound):
    message = _("Store %(store_id)s not enabled in glance.")


class GlanceStoreReadOnly(Invalid):
    message = _("Store %(store_id)s is read-only in glance.")


class VolumeNotFound(NotFound):
    message = _("Volume %(volume_id)s could not be found.")


class MessageNotFound(NotFound):
    message = _("Message %(message_id)s could not be found.")


class VolumeAttachmentNotFound(NotFound):
    message = _("Volume attachment could not be found with "
                "filter: %(filter)s.")


class VolumeMetadataNotFound(NotFound):
    message = _("Volume %(volume_id)s has no metadata with "
                "key %(metadata_key)s.")


class InvalidVolumeMetadata(Invalid):
    message = _("Invalid metadata: %(reason)s")


class InvalidVolumeMetadataSize(Invalid):
    message = _("Invalid metadata size: %(reason)s")


class SnapshotMetadataNotFound(NotFound):
    message = _("Snapshot %(snapshot_id)s has no metadata with "
                "key %(metadata_key)s.")


class VolumeTypeNotFound(NotFound):
    message = _("Volume type %(volume_type_id)s could not be found.")


class VolumeTypeNotFoundByName(VolumeTypeNotFound):
    message = _("Volume type with name %(volume_type_name)s "
                "could not be found.")


class VolumeTypeAccessNotFound(NotFound):
    message = _("Volume type access not found for %(volume_type_id)s / "
                "%(project_id)s combination.")


class VolumeTypeExtraSpecsNotFound(NotFound):
    message = _("Volume Type %(volume_type_id)s has no extra specs with "
                "key %(extra_specs_key)s.")


class VolumeTypeInUse(CinderException):
    message = _("Volume Type %(volume_type_id)s deletion is not allowed with "
                "volumes present with the type.")


class VolumeTypeDeletionError(Invalid):
    message = _("The volume type %(volume_type_id)s is the only currently "
                "defined volume type and cannot be deleted.")


class VolumeTypeDefaultDeletionError(Invalid):
    message = _("The volume type %(volume_type_id)s is a default volume "
                "type and cannot be deleted.")


class VolumeTypeDefaultMisconfiguredError(CinderException):
    message = _("The request cannot be fulfilled as the default volume type "
                "%(volume_type_name)s cannot be found.")


class VolumeTypeProjectDefaultNotFound(NotFound):
    message = _("Default type for project %(project_id)s not found.")


class GroupTypeNotFound(NotFound):
    message = _("Group type %(group_type_id)s could not be found.")


class GroupTypeNotFoundByName(GroupTypeNotFound):
    message = _("Group type with name %(group_type_name)s "
                "could not be found.")


class GroupTypeAccessNotFound(NotFound):
    message = _("Group type access not found for %(group_type_id)s / "
                "%(project_id)s combination.")


class GroupTypeSpecsNotFound(NotFound):
    message = _("Group Type %(group_type_id)s has no specs with "
                "key %(group_specs_key)s.")


class GroupTypeInUse(CinderException):
    message = _("Group Type %(group_type_id)s deletion is not allowed with "
                "groups present with the type.")


class SnapshotNotFound(NotFound):
    message = _("Snapshot %(snapshot_id)s could not be found.")


class ServerNotFound(NotFound):
    message = _("Instance %(uuid)s could not be found.")


class VolumeSnapshotNotFound(NotFound):
    message = _("No snapshots found for volume %(volume_id)s.")


class VolumeIsBusy(CinderException):
    message = _("deleting volume %(volume_name)s that has snapshot")


class SnapshotIsBusy(CinderException):
    message = _("deleting snapshot %(snapshot_name)s that has "
                "dependent volumes")


class InvalidImageRef(Invalid):
    message = _("Invalid image href %(image_href)s.")


class InvalidSignatureImage(Invalid):
    message = _("Signature metadata is incomplete for image: "
                "%(image_id)s.")


class ImageSignatureVerificationException(CinderException):
    message = _("Failed to verify image signature, reason: %(reason)s.")


class ImageNotFound(NotFound):
    message = _("Image %(image_id)s could not be found.")


class ServiceNotFound(NotFound):

    def __init__(self, message=None, **kwargs):
        if not message:
            if kwargs.get('host', None):
                self.message = _("Service %(service_id)s could not be "
                                 "found on host %(host)s.")
            else:
                self.message = _("Service %(service_id)s could not be found.")
        super(ServiceNotFound, self).__init__(message, **kwargs)


class ServiceTooOld(Invalid):
    message = _("Service is too old to fulfil this request.")


class WorkerNotFound(NotFound):
    message = _("Worker with %s could not be found.")

    def __init__(self, message=None, **kwargs):
        keys_list = ('{0}=%({0})s'.format(key) for key in kwargs)
        placeholder = ', '.join(keys_list)
        self.message = self.message % placeholder
        super(WorkerNotFound, self).__init__(message, **kwargs)


class WorkerExists(Duplicate):
    message = _("Worker for %(type)s %(id)s already exists.")


class CleanableInUse(Invalid):
    message = _('%(type)s with id %(id)s is already being cleaned up or '
                'another host has taken over it.')


class ClusterNotFound(NotFound):
    message = _('Cluster %(id)s could not be found.')


class ClusterHasHosts(Invalid):
    message = _("Cluster %(id)s still has hosts.")


class ClusterExists(Duplicate):
    message = _("Cluster %(name)s already exists.")


class HostNotFound(NotFound):
    message = _("Host %(host)s could not be found.")


class SchedulerHostFilterNotFound(NotFound):
    message = _("Scheduler Host Filter %(filter_name)s could not be found.")


class SchedulerHostWeigherNotFound(NotFound):
    message = _("Scheduler Host Weigher %(weigher_name)s could not be found.")


class InvalidReservationExpiration(Invalid):
    message = _("Invalid reservation expiration %(expire)s.")


class InvalidQuotaValue(Invalid):
    message = _("Change would make usage less than 0 for the following "
                "resources: %(unders)s")


class QuotaNotFound(NotFound):
    message = _("Quota could not be found")


class QuotaResourceUnknown(QuotaNotFound):
    message = _("Unknown quota resources %(unknown)s.")


class ProjectQuotaNotFound(QuotaNotFound):
    message = _("Quota for project %(project_id)s could not be found.")


class QuotaClassNotFound(QuotaNotFound):
    message = _("Quota class %(class_name)s could not be found.")


class QuotaUsageNotFound(QuotaNotFound):
    message = _("Quota usage for project %(project_id)s could not be found.")


class OverQuota(CinderException):
    message = _("Quota exceeded for resources: %(overs)s")


class FileNotFound(NotFound):
    message = _("File %(file_path)s could not be found.")


class VolumeTypeExists(Duplicate):
    message = _("Volume Type %(id)s already exists.")


class VolumeTypeAccessExists(Duplicate):
    message = _("Volume type access for %(volume_type_id)s / "
                "%(project_id)s combination already exists.")


class VolumeTypeEncryptionExists(Invalid):
    message = _("Volume type encryption for type %(type_id)s already exists.")


class VolumeTypeEncryptionNotFound(NotFound):
    message = _("Volume type encryption for type %(type_id)s does not exist.")


class GroupTypeExists(Duplicate):
    message = _("Group Type %(id)s already exists.")


class GroupTypeAccessExists(Duplicate):
    message = _("Group type access for %(group_type_id)s / "
                "%(project_id)s combination already exists.")


class GroupVolumeTypeMappingExists(Duplicate):
    message = _("Group volume type mapping for %(group_id)s / "
                "%(volume_type_id)s combination already exists.")


class MalformedRequestBody(CinderException):
    message = _("Malformed message body: %(reason)s")


class ConfigNotFound(NotFound):
    message = _("Could not find config at %(path)s")


class ParameterNotFound(NotFound):
    message = _("Could not find parameter %(param)s")


class NoValidBackend(CinderException):
    message = _("No valid backend was found. %(reason)s")


class QuotaError(CinderException):
    message = _("Quota exceeded: code=%(code)s")
    code = 413
    headers = {'Retry-After': '0'}
    safe = True


class VolumeSizeExceedsAvailableQuota(QuotaError):
    message = _("Requested volume or snapshot exceeds allowed %(name)s "
                "quota. Requested %(requested)sG, quota is %(quota)sG and "
                "%(consumed)sG has been consumed.")

    def __init__(self, message=None, **kwargs):
        kwargs.setdefault('name', 'gigabytes')
        super(VolumeSizeExceedsAvailableQuota, self).__init__(
            message, **kwargs)


class VolumeSizeExceedsLimit(QuotaError):
    message = _("Requested volume size %(size)dG is larger than "
                "maximum allowed limit %(limit)dG.")


class VolumeBackupSizeExceedsAvailableQuota(QuotaError):
    message = _("Requested backup exceeds allowed Backup gigabytes "
                "quota. Requested %(requested)sG, quota is %(quota)sG and "
                "%(consumed)sG has been consumed.")


class VolumeLimitExceeded(QuotaError):
    message = _("Maximum number of volumes allowed (%(allowed)d) exceeded for "
                "quota '%(name)s'.")

    def __init__(self, message=None, **kwargs):
        kwargs.setdefault('name', 'volumes')
        super(VolumeLimitExceeded, self).__init__(message, **kwargs)


class SnapshotLimitExceeded(QuotaError):
    message = _("Maximum number of snapshots allowed (%(allowed)d) exceeded")


class UnexpectedOverQuota(QuotaError):
    message = _("Unexpected over quota on %(name)s.")


class BackupLimitExceeded(QuotaError):
    message = _("Maximum number of backups allowed (%(allowed)d) exceeded")


class ImageLimitExceeded(QuotaError):
    message = _("Image quota exceeded")


class VolumeTypeCreateFailed(CinderException):
    message = _("Cannot create volume_type with "
                "name %(name)s and specs %(extra_specs)s")


class VolumeTypeUpdateFailed(CinderException):
    message = _("Cannot update volume_type %(id)s")


class GroupTypeCreateFailed(CinderException):
    message = _("Cannot create group_type with "
                "name %(name)s and specs %(group_specs)s")


class GroupTypeUpdateFailed(CinderException):
    message = _("Cannot update group_type %(id)s")


class GroupLimitExceeded(QuotaError):
    message = _("Maximum number of groups allowed (%(allowed)d) exceeded")


class UnknownCmd(VolumeDriverException):
    message = _("Unknown or unsupported command %(cmd)s")


class MalformedResponse(VolumeDriverException):
    message = _("Malformed response to command %(cmd)s: %(reason)s")


class FailedCmdWithDump(VolumeDriverException):
    message = _("Operation failed with status=%(status)s. Full dump: %(data)s")


class InvalidConnectorException(VolumeDriverException):
    message = _("Connector doesn't have required information: %(missing)s")


class GlanceMetadataExists(Invalid):
    message = _("Glance metadata cannot be updated, key %(key)s"
                " exists for volume id %(volume_id)s")


class GlanceMetadataNotFound(NotFound):
    message = _("Glance metadata for volume/snapshot %(id)s cannot be found.")


class ImageDownloadFailed(CinderException):
    message = _("Failed to download image %(image_href)s, reason: %(reason)s")


class ExportFailure(Invalid):
    message = _("Failed to export for volume: %(reason)s")


class RemoveExportException(VolumeDriverException):
    message = _("Failed to remove export for volume %(volume)s: %(reason)s")


class MetadataUpdateFailure(Invalid):
    message = _("Failed to update metadata for volume: %(reason)s")


class MetadataCopyFailure(Invalid):
    message = _("Failed to copy metadata to volume: %(reason)s")


class InvalidMetadataType(Invalid):
    message = _("The type of metadata: %(metadata_type)s for volume/snapshot "
                "%(id)s is invalid.")


class ImageCopyFailure(Invalid):
    message = _("Failed to copy image to volume: %(reason)s")


class BackupInvalidCephArgs(BackupDriverException):
    message = _("Invalid Ceph args provided for backup rbd operation")


class BackupOperationError(Invalid):
    message = _("An error has occurred during backup operation")


class BackupMetadataUnsupportedVersion(BackupDriverException):
    message = _("Unsupported backup metadata version requested")


class BackupMetadataNotFound(NotFound):
    message = _("Backup %(backup_id)s has no metadata with "
                "key %(metadata_key)s.")


class VolumeMetadataBackupExists(BackupDriverException):
    message = _("Metadata backup already exists for this volume")


class BackupRBDOperationFailed(BackupDriverException):
    message = _("Backup RBD operation failed")


class EncryptedBackupOperationFailed(BackupDriverException):
    message = _("Backup operation of an encrypted volume failed.")


class BackupNotFound(NotFound):
    message = _("Backup %(backup_id)s could not be found.")


class InvalidBackup(Invalid):
    message = _("Invalid backup: %(reason)s")


class SwiftConnectionFailed(BackupDriverException):
    message = _("Connection to swift failed: %(reason)s")


class TransferNotFound(NotFound):
    message = _("Transfer %(transfer_id)s could not be found.")


class VolumeMigrationFailed(CinderException):
    message = _("Volume migration failed: %(reason)s")


class SSHInjectionThreat(CinderException):
    message = _("SSH command injection detected: %(command)s")


class QoSSpecsExists(Duplicate):
    message = _("QoS Specs %(specs_id)s already exists.")


class QoSSpecsCreateFailed(CinderException):
    message = _("Failed to create qos_specs: "
                "%(name)s with specs %(qos_specs)s.")


class QoSSpecsUpdateFailed(CinderException):
    message = _("Failed to update qos_specs: "
                "%(specs_id)s with specs %(qos_specs)s.")


class QoSSpecsNotFound(NotFound):
    message = _("No such QoS spec %(specs_id)s.")


class QoSSpecsAssociateFailed(CinderException):
    message = _("Failed to associate qos_specs: "
                "%(specs_id)s with type %(type_id)s.")


class QoSSpecsDisassociateFailed(CinderException):
    message = _("Failed to disassociate qos_specs: "
                "%(specs_id)s with type %(type_id)s.")


class QoSSpecsKeyNotFound(NotFound):
    message = _("QoS spec %(specs_id)s has no spec with "
                "key %(specs_key)s.")


class InvalidQoSSpecs(Invalid):
    message = _("Invalid qos specs: %(reason)s")


class QoSSpecsInUse(CinderException):
    message = _("QoS Specs %(specs_id)s is still associated with entities.")


class KeyManagerError(CinderException):
    message = _("key manager error: %(reason)s")


class ManageExistingInvalidReference(CinderException):
    message = _("Manage existing volume failed due to invalid backend "
                "reference %(existing_ref)s: %(reason)s")


class ManageExistingAlreadyManaged(CinderException):
    message = _("Unable to manage existing volume. "
                "Volume %(volume_ref)s already managed.")


class InvalidReplicationTarget(Invalid):
    message = _("Invalid Replication Target: %(reason)s")


class UnableToFailOver(CinderException):
    message = _("Unable to failover to replication target: %(reason)s).")


class ReplicationError(CinderException):
    message = _("Volume %(volume_id)s replication "
                "error: %(reason)s")


class ReplicationGroupError(CinderException):
    message = _("Group %(group_id)s replication "
                "error: %(reason)s.")


class ManageExistingVolumeTypeMismatch(CinderException):
    message = _("Manage existing volume failed due to volume type mismatch: "
                "%(reason)s")


class ExtendVolumeError(CinderException):
    message = _("Error extending volume: %(reason)s")


class EvaluatorParseException(Exception):
    message = _("Error during evaluator parsing: %(reason)s")


class LockCreationFailed(CinderException):
    message = _('Unable to create lock. Coordination backend not started.')


OrphanedObjectError = obj_exc.OrphanedObjectError
ObjectActionError = obj_exc.ObjectActionError


class CappedVersionUnknown(CinderException):
    message = _("Unrecoverable Error: Versioned Objects in DB are capped to "
                "unknown version %(version)s. Most likely your environment "
                "contains only new services and you're trying to start an "
                "older one. Use `cinder-manage service list` to check that "
                "and upgrade this service.")


class VolumeGroupNotFound(CinderException):
    message = _('Unable to find Volume Group: %(vg_name)s')


class VolumeGroupCreationFailed(CinderException):
    message = _('Failed to create Volume Group: %(vg_name)s')


class VolumeNotDeactivated(CinderException):
    message = _('Volume %(name)s was not deactivated in time.')


class VolumeDeviceNotFound(CinderException):
    message = _('Volume device not found at %(device)s.')


# RemoteFS drivers
class RemoteFSException(VolumeDriverException):
    message = _("Unknown RemoteFS exception")


class RemoteFSConcurrentRequest(RemoteFSException):
    message = _("A concurrent, possibly contradictory, request "
                "has been made.")


class RemoteFSNoSharesMounted(RemoteFSException):
    message = _("No mounted shares found")


class RemoteFSNoSuitableShareFound(RemoteFSException):
    message = _("There is no share which can host %(volume_size)sG")


class RemoteFSInvalidBackingFile(VolumeDriverException):
    message = _("File %(path)s has invalid backing file %(backing_file)s.")


# NFS driver
class NfsException(RemoteFSException):
    message = _("Unknown NFS exception")


class NfsNoSharesMounted(RemoteFSNoSharesMounted):
    message = _("No mounted NFS shares found")


class NfsNoSuitableShareFound(RemoteFSNoSuitableShareFound):
    message = _("There is no share which can host %(volume_size)sG")


# Fibre Channel Zone Manager
class ZoneManagerException(CinderException):
    message = _("Fibre Channel connection control failure: %(reason)s")


class FCZoneDriverException(CinderException):
    message = _("Fibre Channel Zone operation failed: %(reason)s")


class FCSanLookupServiceException(CinderException):
    message = _("Fibre Channel SAN Lookup failure: %(reason)s")


class ZoneManagerNotInitialized(CinderException):
    message = _("Fibre Channel Zone Manager not initialized")


# ConsistencyGroup
class ConsistencyGroupNotFound(NotFound):
    message = _("ConsistencyGroup %(consistencygroup_id)s could not be found.")


class InvalidConsistencyGroup(Invalid):
    message = _("Invalid ConsistencyGroup: %(reason)s")


# Group
class GroupNotFound(NotFound):
    message = _("Group %(group_id)s could not be found.")


class InvalidGroup(Invalid):
    message = _("Invalid Group: %(reason)s")


class InvalidGroupStatus(Invalid):
    message = _("Invalid Group Status: %(reason)s")


# CgSnapshot
class CgSnapshotNotFound(NotFound):
    message = _("CgSnapshot %(cgsnapshot_id)s could not be found.")


class InvalidCgSnapshot(Invalid):
    message = _("Invalid CgSnapshot: %(reason)s")


# GroupSnapshot
class GroupSnapshotNotFound(NotFound):
    message = _("GroupSnapshot %(group_snapshot_id)s could not be found.")


class InvalidGroupSnapshot(Invalid):
    message = _("Invalid GroupSnapshot: %(reason)s")


class InvalidGroupSnapshotStatus(Invalid):
    message = _("Invalid GroupSnapshot Status: %(reason)s")


# Target drivers
class ISCSITargetCreateFailed(CinderException):
    message = _("Failed to create iscsi target for volume %(volume_id)s.")


class ISCSITargetRemoveFailed(CinderException):
    message = _("Failed to remove iscsi target for volume %(volume_id)s.")


class ISCSITargetAttachFailed(CinderException):
    message = _("Failed to attach iSCSI target for volume %(volume_id)s.")


class ISCSITargetDetachFailed(CinderException):
    message = _("Failed to detach iSCSI target for volume %(volume_id)s.")


class TargetUpdateFailed(CinderException):
    message = _("Failed to update target for volume %(volume_id)s.")


class ISCSITargetHelperCommandFailed(CinderException):
    message = "%(error_message)s"


class BadHTTPResponseStatus(VolumeDriverException):
    message = _("Bad HTTP response status %(status)s")


class BadResetResourceStatus(CinderException):
    message = _("Bad reset resource status : %(reason)s")


class MetadataAbsent(CinderException):
    message = _("There is no metadata in DB object.")


class NotSupportedOperation(Invalid):
    message = _("Operation not supported: %(operation)s.")
    code = 405


class AttachmentSpecsNotFound(NotFound):
    message = _("Attachment %(attachment_id)s has no "
                "key %(specs_key)s.")


class InvalidName(Invalid):
    message = _("An invalid 'name' value was provided. %(reason)s")


class ServiceUserTokenNoAuth(CinderException):
    message = _("The [service_user] send_service_user_token option was "
                "requested, but no service auth could be loaded. Please check "
                "the [service_user] configuration section.")


class RekeyNotSupported(CinderException):
    message = _("Rekey not supported.")


class ImageCompressionNotAllowed(CinderException):
    message = _("Image compression upload disallowed, but container_format "
                "is compressed")


class ImageConversionNotAllowed(CinderException):
    message = _("Image Conversion disallowed for image %(image_id)s: "
                "%(reason)s")


class CinderAcceleratorError(CinderException):
    message = _("Cinder accelerator %(accelerator)s encountered an error "
                "while compressing/decompressing image.\n"
                "Command %(cmd)s execution failed.\n"
                "%(description)s\n"
                "Reason: %(reason)s")


class SnapshotLimitReached(CinderException):
    message = _("Exceeded the configured limit of "
                "%(set_limit)s snapshots per volume.")


class DriverInitiatorDataExists(Duplicate):
    message = _(
        "Driver initiator data for initiator '%(initiator)s' and backend "
        "'%(namespace)s' with key '%(key)s' already exists."
    )



# Copyright 2014 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Tests For Volume Number Weigher.
"""

from unittest import mock

from cinder.common import constants
from cinder import context
from cinder.db.sqlalchemy import api
from cinder.scheduler import weights
from cinder.tests.unit import fake_constants
from cinder.tests.unit.scheduler import fakes
from cinder.tests.unit import test
from cinder.volume import volume_utils


def fake_volume_data_get_for_host(context, host, count_only=False):
    host = volume_utils.extract_host(host)
    if host == 'host1':
        return 1
    elif host == 'host2':
        return 2
    elif host == 'host3':
        return 3
    elif host == 'host4':
        return 4
    elif host == 'host5':
        return 5
    else:
        return 6


class VolumeNumberWeigherTestCase(test.TestCase):

    def setUp(self):
        super(VolumeNumberWeigherTestCase, self).setUp()
        uid = fake_constants.USER_ID
        pid = fake_constants.PROJECT_ID
        self.context = context.RequestContext(user_id=uid,
                                              project_id=pid,
                                              is_admin=False,
                                              read_deleted="no",
                                              overwrite=False)
        self.host_manager = fakes.FakeHostManager()
        self.weight_handler = weights.OrderedHostWeightHandler(
            'cinder.scheduler.weights')

    def _get_weighed_host(self, hosts, weight_properties=None):
        if weight_properties is None:
            weight_properties = {'context': self.context}
        return self.weight_handler.get_weighed_objects(
            [weights.volume_number.VolumeNumberWeigher],
            hosts,
            weight_properties)[0]

    @mock.patch('cinder.db.sqlalchemy.api.service_get_all')
    def _get_all_backends(self, _mock_service_get_all, disabled=False):
        ctxt = context.get_admin_context()
        fakes.mock_host_manager_db_calls(_mock_service_get_all,
                                         disabled=disabled)
        backend_states = self.host_manager.get_all_backend_states(ctxt)
        _mock_service_get_all.assert_called_once_with(
            ctxt,
            None,  # backend_match_level
            topic=constants.VOLUME_TOPIC,
            frozen=False,
            disabled=disabled)
        return backend_states

    def test_volume_number_weight_multiplier1(self):
        self.flags(volume_number_multiplier=-1.0)
        backend_info_list = self._get_all_backends()

        # host1: 1 volume    Norm=0.0
        # host2: 2 volumes
        # host3: 3 volumes
        # host4: 4 volumes
        # host5: 5 volumes   Norm=-1.0
        # so, host1 should win:
        with mock.patch.object(api, 'volume_data_get_for_host',
                               fake_volume_data_get_for_host):
            weighed_host = self._get_weighed_host(backend_info_list)
            self.assertEqual(0.0, weighed_host.weight)
            self.assertEqual('host1',
                             volume_utils.extract_host(weighed_host.obj.host))

    def test_volume_number_weight_multiplier2(self):
        self.flags(volume_number_multiplier=1.0)
        backend_info_list = self._get_all_backends()

        # host1: 1 volume      Norm=0
        # host2: 2 volumes
        # host3: 3 volumes
        # host4: 4 volumes
        # host5: 5 volumes     Norm=1
        # so, host5 should win:
        with mock.patch.object(api, 'volume_data_get_for_host',
                               fake_volume_data_get_for_host):
            weighed_host = self._get_weighed_host(backend_info_list)
            self.assertEqual(1.0, weighed_host.weight)
            self.assertEqual('host5',
                             volume_utils.extract_host(weighed_host.obj.host))



#    Copyright 2015 SimpliVity Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import versionutils
from oslo_versionedobjects import fields

from cinder import db
from cinder import exception
from cinder.i18n import _
from cinder import objects
from cinder.objects import base
from cinder.objects import cleanable
from cinder.objects import fields as c_fields
from cinder.volume import volume_types

CONF = cfg.CONF

LOG = logging.getLogger(__name__)


class MetadataObject(dict):
    # This is a wrapper class that simulates SQLAlchemy (.*)Metadata objects to
    # maintain compatibility with older representations of Volume that some
    # drivers rely on. This is helpful in transition period while some driver
    # methods are invoked with volume versioned object and some SQLAlchemy
    # object or dict.
    def __init__(self, key=None, value=None):
        super(MetadataObject, self).__init__()
        self.key = key
        self.value = value

    def __getattr__(self, name):
        if name in self:
            return self[name]
        else:
            raise AttributeError("No such attribute: " + name)

    def __setattr__(self, name, value):
        self[name] = value


@base.CinderObjectRegistry.register
class Volume(cleanable.CinderCleanableObject, base.CinderObject,
             base.CinderObjectDictCompat, base.CinderComparableObject,
             base.ClusteredObject):
    # Version 1.0: Initial version
    # Version 1.1: Added metadata, admin_metadata, volume_attachment, and
    #              volume_type
    # Version 1.2: Added glance_metadata, consistencygroup and snapshots
    # Version 1.3: Added finish_volume_migration()
    # Version 1.4: Added cluster fields
    # Version 1.5: Added group
    # Version 1.6: This object is now cleanable (adds rows to workers table)
    # Version 1.7: Added service_uuid
    # Version 1.8: Added shared_targets
    # Version 1.9: Added use_quota
    VERSION = '1.9'

    OPTIONAL_FIELDS = ('metadata', 'admin_metadata', 'glance_metadata',
                       'volume_type', 'volume_attachment', 'consistencygroup',
                       'snapshots', 'cluster', 'group')

    # NOTE: When adding a field obj_make_compatible needs to be updated
    fields = {
        # id is the user facing UUID that should be passed to API calls
        'id': fields.UUIDField(),
        # _name_id is the real volume's UUID that should be used by the driver
        # when it is set.  This is used when migrating a volume.  Property
        # name_id is provided for convenience.
        '_name_id': fields.UUIDField(nullable=True),
        'ec2_id': fields.UUIDField(nullable=True),
        'user_id': fields.StringField(nullable=True),
        'project_id': fields.StringField(nullable=True),

        # TODO: (Y release) Change nullable to False
        'use_quota': fields.BooleanField(default=True, nullable=True),
        'snapshot_id': fields.UUIDField(nullable=True),

        'cluster_name': fields.StringField(nullable=True),
        'cluster': fields.ObjectField('Cluster', nullable=True,
                                      read_only=True),
        'host': fields.StringField(nullable=True),
        'size': fields.IntegerField(nullable=True),
        'availability_zone': fields.StringField(nullable=True),
        'status': fields.StringField(nullable=True),
        'attach_status': c_fields.VolumeAttachStatusField(nullable=True),
        'migration_status': fields.StringField(nullable=True),

        'scheduled_at': fields.DateTimeField(nullable=True),
        'launched_at': fields.DateTimeField(nullable=True),
        'terminated_at': fields.DateTimeField(nullable=True),

        'display_name': fields.StringField(nullable=True),
        'display_description': fields.StringField(nullable=True),

        'provider_id': fields.StringField(nullable=True),
        'provider_location': fields.StringField(nullable=True),
        'provider_auth': fields.StringField(nullable=True),
        'provider_geometry': fields.StringField(nullable=True),

        'volume_type_id': fields.UUIDField(nullable=True),
        'source_volid': fields.UUIDField(nullable=True),
        'encryption_key_id': fields.UUIDField(nullable=True),

        'consistencygroup_id': fields.UUIDField(nullable=True),
        'group_id': fields.UUIDField(nullable=True),

        'deleted': fields.BooleanField(default=False, nullable=True),
        'bootable': fields.BooleanField(default=False, nullable=True),
        'multiattach': fields.BooleanField(default=False, nullable=True),

        'replication_status': fields.StringField(nullable=True),
        'replication_extended_status': fields.StringField(nullable=True),
        'replication_driver_data': fields.StringField(nullable=True),

        'previous_status': fields.StringField(nullable=True),

        'metadata': fields.DictOfStringsField(nullable=True),
        'admin_metadata': fields.DictOfStringsField(nullable=True),
        'glance_metadata': fields.DictOfStringsField(nullable=True),
        'volume_type': fields.ObjectField('VolumeType', nullable=True),
        'volume_attachment': fields.ObjectField('VolumeAttachmentList',
                                                nullable=True),
        'consistencygroup': fields.ObjectField('ConsistencyGroup',
                                               nullable=True),
        'snapshots': fields.ObjectField('SnapshotList', nullable=True),
        'group': fields.ObjectField('Group', nullable=True),
        'service_uuid': fields.StringField(nullable=True),
        'shared_targets': fields.BooleanField(default=True, nullable=True),
    }

    # NOTE(thangp): obj_extra_fields is used to hold properties that are not
    # usually part of the model
    obj_extra_fields = ['name', 'name_id', 'volume_metadata',
                        'volume_admin_metadata', 'volume_glance_metadata']

    @classmethod
    def _get_expected_attrs(cls, context, *args, **kwargs):
        expected_attrs = ['metadata', 'volume_type', 'volume_type.extra_specs',
                          'volume_attachment']
        if context.is_admin:
            expected_attrs.append('admin_metadata')

        return expected_attrs

    @property
    def name_id(self):
        """Actual volume's UUID for driver usage.

        There may be two different UUIDs for the same volume, the user facing
        one, and the one the driver should be using.

        When a volume is created these two are the same, but when doing a
        generic migration (create new volume, then copying data) they will be
        different if we were unable to rename the new volume in the final
        migration steps.

        So the volume will have been created using the new volume's UUID and
        the driver will have to look for it using that UUID, but the user on
        the other hand will keep referencing the volume with the original UUID.

        This property facilitates using the right UUID in the driver's code.
        """
        return self.id if not self._name_id else self._name_id

    @name_id.setter
    def name_id(self, value):
        self._name_id = value

    @property
    def name(self):
        return CONF.volume_name_template % self.name_id

    # TODO(dulek): Three properties below are for compatibility with dict
    # representation of volume. The format there is different (list of
    # SQLAlchemy models) so we need a conversion. Anyway - these should be
    # removed when we stop this class from deriving from DictObjectCompat.
    @property
    def volume_metadata(self):
        md = [MetadataObject(k, v) for k, v in self.metadata.items()]
        return md

    @volume_metadata.setter
    def volume_metadata(self, value):
        md = {d['key']: d['value'] for d in value}
        self.metadata = md

    @property
    def volume_admin_metadata(self):
        md = [MetadataObject(k, v) for k, v in self.admin_metadata.items()]
        return md

    @volume_admin_metadata.setter
    def volume_admin_metadata(self, value):
        md = {d['key']: d['value'] for d in value}
        self.admin_metadata = md

    def admin_metadata_update(self, metadata, delete, add=True, update=True):
        new_metadata = db.volume_admin_metadata_update(self._context, self.id,
                                                       metadata, delete, add,
                                                       update)
        self.admin_metadata = new_metadata
        self._reset_metadata_tracking(fields=('admin_metadata',))

    @property
    def volume_glance_metadata(self):
        md = [MetadataObject(k, v) for k, v in self.glance_metadata.items()]
        return md

    @volume_glance_metadata.setter
    def volume_glance_metadata(self, value):
        md = {d['key']: d['value'] for d in value}
        self.glance_metadata = md

    def __init__(self, *args, **kwargs):
        super(Volume, self).__init__(*args, **kwargs)

        self._reset_metadata_tracking()

    def obj_reset_changes(self, fields=None):
        super(Volume, self).obj_reset_changes(fields)
        self._reset_metadata_tracking(fields=fields)

    @classmethod
    def _obj_from_primitive(cls, context, objver, primitive):
        # TODO: (Y release) remove next line
        cls._ensure_use_quota_is_set(primitive['versioned_object.data'])
        obj = super(Volume, Volume)._obj_from_primitive(context, objver,
                                                        primitive)
        obj._reset_metadata_tracking()
        return obj

    def _reset_metadata_tracking(self, fields=None):
        if fields is None or 'metadata' in fields:
            self._orig_metadata = (dict(self.metadata)
                                   if 'metadata' in self else {})
        if fields is None or 'admin_metadata' in fields:
            self._orig_admin_metadata = (dict(self.admin_metadata)
                                         if 'admin_metadata' in self
                                         else {})
        if fields is None or 'glance_metadata' in fields:
            self._orig_glance_metadata = (dict(self.glance_metadata)
                                          if 'glance_metadata' in self
                                          else {})

    def obj_what_changed(self):
        changes = super(Volume, self).obj_what_changed()
        if 'metadata' in self and self.metadata != self._orig_metadata:
            changes.add('metadata')
        if ('admin_metadata' in self and
                self.admin_metadata != self._orig_admin_metadata):
            changes.add('admin_metadata')
        if ('glance_metadata' in self and
                self.glance_metadata != self._orig_glance_metadata):
            changes.add('glance_metadata')

        return changes

    def obj_make_compatible(self, primitive, target_version):
        """Make a Volume representation compatible with a target version."""
        super(Volume, self).obj_make_compatible(primitive, target_version)
        target_version = versionutils.convert_version_to_tuple(target_version)
        # TODO: (Y release) remove next 2 lines & method if nothing else below
        if target_version < (1, 9):
            primitive.pop('use_quota', None)

    @classmethod
    def _from_db_object(cls, context, volume, db_volume, expected_attrs=None):
        if expected_attrs is None:
            expected_attrs = []
        for name, field in volume.fields.items():
            if name in cls.OPTIONAL_FIELDS:
                continue
            value = db_volume.get(name)
            if isinstance(field, fields.IntegerField):
                value = value or 0
            volume[name] = value

        # Get data from db_volume object that was queried by joined query
        # from DB
        if 'metadata' in expected_attrs:
            metadata = db_volume.get('volume_metadata', [])
            volume.metadata = {item['key']: item['value'] for item in metadata}
        if 'admin_metadata' in expected_attrs:
            metadata = db_volume.get('volume_admin_metadata', [])
            volume.admin_metadata = {item['key']: item['value']
                                     for item in metadata}
        if 'glance_metadata' in expected_attrs:
            metadata = db_volume.get('volume_glance_metadata', [])
            volume.glance_metadata = {item['key']: item['value']
                                      for item in metadata}
        if 'volume_type' in expected_attrs:
            db_volume_type = db_volume.get('volume_type')
            if db_volume_type:
                vt_expected_attrs = []
                if 'volume_type.extra_specs' in expected_attrs:
                    vt_expected_attrs.append('extra_specs')
                volume.volume_type = objects.VolumeType._from_db_object(
                    context, objects.VolumeType(), db_volume_type,
                    expected_attrs=vt_expected_attrs)
        if 'volume_attachment' in expected_attrs:
            attachments = base.obj_make_list(
                context, objects.VolumeAttachmentList(context),
                objects.VolumeAttachment,
                db_volume.get('volume_attachment'))
            volume.volume_attachment = attachments
        if volume.consistencygroup_id and 'consistencygroup' in expected_attrs:
            consistencygroup = objects.ConsistencyGroup(context)
            consistencygroup._from_db_object(context,
                                             consistencygroup,
                                             db_volume['consistencygroup'])
            volume.consistencygroup = consistencygroup
        if 'snapshots' in expected_attrs:
            snapshots = base.obj_make_list(
                context, objects.SnapshotList(context),
                objects.Snapshot,
                db_volume['snapshots'])
            volume.snapshots = snapshots
        if 'cluster' in expected_attrs:
            db_cluster = db_volume.get('cluster')
            # If this volume doesn't belong to a cluster the cluster field in
            # the ORM instance will have value of None.
            if db_cluster:
                volume.cluster = objects.Cluster(context)
                objects.Cluster._from_db_object(context, volume.cluster,
                                                db_cluster)
            else:
                volume.cluster = None
        if volume.group_id and 'group' in expected_attrs:
            group = objects.Group(context)
            group._from_db_object(context,
                                  group,
                                  db_volume['group'])
            volume.group = group

        volume._context = context
        volume.obj_reset_changes()
        return volume

    # TODO: (Z release): Remove method and leave the default of False from DB
    @staticmethod
    def _ensure_use_quota_is_set(updates, warning=False):
        if updates.get('use_quota') is None:
            use_quota = not (
                (updates.get('migration_status') or ''
                 ).startswith('target:') or
                (updates.get('admin_metadata') or {}
                 ).get('temporary') == 'True')
            if warning and not use_quota:
                LOG.warning('Ooooops, we forgot to set the use_quota field to '
                            'False!!  Fix code here')
            updates['use_quota'] = use_quota

    def populate_consistencygroup(self):
        """Populate CG fields based on group fields.

        Method assumes that consistencygroup_id and consistencygroup fields
        have not already been set.

        This is a hack to support backward compatibility of consistencygroup,
        where we set the fields but don't want to write them to the DB, so we
        mark them as not changed, so they won't be stored on the next save().
        """
        self.consistencygroup_id = self.group_id
        if self.group_id and self.obj_attr_is_set('group'):
            cg = objects.ConsistencyGroup()
            cg.from_group(self.group)
            self.consistencygroup = cg
        self.obj_reset_changes(['consistencygroup', 'consistencygroup_id'])

    def create(self):
        if self.obj_attr_is_set('id'):
            raise exception.ObjectActionError(action='create',
                                              reason=_('already created'))
        updates = self.cinder_obj_get_changes()

        if 'consistencygroup' in updates:
            raise exception.ObjectActionError(
                action='create', reason=_('consistencygroup assigned'))
        if 'snapshots' in updates:
            raise exception.ObjectActionError(
                action='create', reason=_('snapshots assigned'))
        if 'cluster' in updates:
            raise exception.ObjectActionError(
                action='create', reason=_('cluster assigned'))
        if 'group' in updates:
            raise exception.ObjectActionError(
                action='create', reason=_('group assigned'))
        if ('volume_type_id' not in updates or
                updates['volume_type_id'] is None):
            updates['volume_type_id'] = (
                volume_types.get_default_volume_type()['id'])

        # TODO: (Y release) Remove this call since we should have already made
        # all methods in Cinder make the call with the right values.
        self._ensure_use_quota_is_set(updates, warning=True)

        db_volume = db.volume_create(self._context, updates)
        expected_attrs = self._get_expected_attrs(self._context)
        self._from_db_object(self._context, self, db_volume, expected_attrs)

    def save(self):
        # TODO: (Y release) Remove this online migration code
        # Pass self directly since it's a CinderObjectDictCompat
        self._ensure_use_quota_is_set(self)

        updates = self.cinder_obj_get_changes()
        if updates:
            # NOTE(xyang): Allow this to pass if 'consistencygroup' is
            # set to None. This is to support backward compatibility.
            # Also remove 'consistencygroup' from updates because
            # consistencygroup is the name of a relationship in the ORM
            # Volume model, so SQLA tries to do some kind of update of
            # the foreign key based on the provided updates if
            # 'consistencygroup' is in updates.
            if updates.pop('consistencygroup', None):
                raise exception.ObjectActionError(
                    action='save', reason=_('consistencygroup changed'))
            if 'group' in updates:
                raise exception.ObjectActionError(
                    action='save', reason=_('group changed'))
            if 'glance_metadata' in updates:
                raise exception.ObjectActionError(
                    action='save', reason=_('glance_metadata changed'))
            if 'snapshots' in updates:
                raise exception.ObjectActionError(
                    action='save', reason=_('snapshots changed'))
            if 'cluster' in updates:
                raise exception.ObjectActionError(
                    action='save', reason=_('cluster changed'))
            if 'metadata' in updates:
                # Metadata items that are not specified in the
                # self.metadata will be deleted
                metadata = updates.pop('metadata', None)
                self.metadata = db.volume_metadata_update(self._context,
                                                          self.id, metadata,
                                                          True)
            if self._context.is_admin and 'admin_metadata' in updates:
                metadata = updates.pop('admin_metadata', None)
                self.admin_metadata = db.volume_admin_metadata_update(
                    self._context, self.id, metadata, True)

            # When we are creating a volume and we change from 'creating'
            # status to 'downloading' status we have to change the worker entry
            # in the DB to reflect this change, otherwise the cleanup will
            # not be performed as it will be mistaken for a volume that has
            # been somehow changed (reset status, forced operation...)
            if updates.get('status') == 'downloading':
                self.set_worker()

            # updates are changed after popping out metadata.
            if updates:
                db.volume_update(self._context, self.id, updates)
            self.obj_reset_changes()

    def destroy(self):
        with self.obj_as_admin():
            updated_values = db.volume_destroy(self._context, self.id)
        self.update(updated_values)
        self.obj_reset_changes(updated_values.keys())

    def obj_load_attr(self, attrname):
        if attrname not in self.OPTIONAL_FIELDS:
            raise exception.ObjectActionError(
                action='obj_load_attr',
                reason=_('attribute %s not lazy-loadable') % attrname)
        if not self._context:
            raise exception.OrphanedObjectError(method='obj_load_attr',
                                                objtype=self.obj_name())

        if attrname == 'metadata':
            self.metadata = db.volume_metadata_get(self._context, self.id)
        elif attrname == 'admin_metadata':
            self.admin_metadata = {}
            if self._context.is_admin:
                self.admin_metadata = db.volume_admin_metadata_get(
                    self._context, self.id)
        elif attrname == 'glance_metadata':
            try:
                # NOTE(dulek): We're using alias here to have conversion from
                # list to dict done there.
                self.volume_glance_metadata = db.volume_glance_metadata_get(
                    self._context, self.id)
            except exception.GlanceMetadataNotFound:
                # NOTE(dulek): DB API raises when volume has no
                # glance_metadata. Silencing this because at this level no
                # metadata is a completely valid result.
                self.glance_metadata = {}
        elif attrname == 'volume_type':
            # If the volume doesn't have volume_type, VolumeType.get_by_id
            # would trigger a db call which raise VolumeTypeNotFound exception.
            self.volume_type = (objects.VolumeType.get_by_id(
                self._context, self.volume_type_id) if self.volume_type_id
                else None)
        elif attrname == 'volume_attachment':
            attachments = objects.VolumeAttachmentList.get_all_by_volume_id(
                self._context, self.id)
            self.volume_attachment = attachments
        elif attrname == 'consistencygroup':
            if self.consistencygroup_id is None:
                self.consistencygroup = None
            else:
                consistencygroup = objects.ConsistencyGroup.get_by_id(
                    self._context, self.consistencygroup_id)
                self.consistencygroup = consistencygroup
        elif attrname == 'snapshots':
            self.snapshots = objects.SnapshotList.get_all_for_volume(
                self._context, self.id)
        elif attrname == 'cluster':
            # If this volume doesn't belong to a cluster (cluster_name is
            # empty), then cluster field will be None.
            if self.cluster_name:
                self.cluster = objects.Cluster.get_by_id(
                    self._context, name=self.cluster_name)
            else:
                self.cluster = None
        elif attrname == 'group':
            if self.group_id is None:
                self.group = None
            else:
                group = objects.Group.get_by_id(
                    self._context, self.group_id)
                self.group = group

        self.obj_reset_changes(fields=[attrname])

    def delete_metadata_key(self, key):
        db.volume_metadata_delete(self._context, self.id, key)
        md_was_changed = 'metadata' in self.obj_what_changed()

        del self.metadata[key]
        self._orig_metadata.pop(key, None)

        if not md_was_changed:
            self.obj_reset_changes(['metadata'])

    def finish_volume_migration(self, dest_volume):
        # We swap fields between source (i.e. self) and destination at the
        # end of migration because we want to keep the original volume id
        # in the DB but now pointing to the migrated volume.
        skip = ({'id', 'provider_location', 'glance_metadata', 'use_quota',
                 'volume_type', 'volume_attachment'}
                | set(self.obj_extra_fields))
        for key in set(dest_volume.fields.keys()) - skip:
            # Only swap attributes that are already set.  We do not want to
            # unexpectedly trigger a lazy-load.
            if not dest_volume.obj_attr_is_set(key):
                continue

            value = getattr(dest_volume, key)
            value_to_dst = getattr(self, key)

            # Destination must have a _name_id since the id no longer matches
            # the volume.  If it doesn't have a _name_id we set one.
            if key == '_name_id':
                if not dest_volume._name_id:
                    setattr(dest_volume, key, self.id)
                continue
            elif key == 'migration_status':
                value = None
                value_to_dst = 'deleting'
            elif key == 'display_description':
                value_to_dst = 'migration src for ' + self.id
            elif key == 'status':
                value_to_dst = 'deleting'
            # Because dest_volume will be deleted soon, we can
            # skip to copy volume_type_id and volume_type which
            # are not keys for volume deletion.
            elif key == 'volume_type_id':
                # Initialize volume_type of source volume using
                # new volume_type_id.
                self.update({'volume_type_id': value})
                continue

            setattr(self, key, value)
            setattr(dest_volume, key, value_to_dst)

        self.save()
        dest_volume.save()
        return dest_volume

    def get_latest_snapshot(self):
        """Get volume's latest snapshot"""
        snapshot_db = db.snapshot_get_latest_for_volume(self._context, self.id)
        snapshot = objects.Snapshot(self._context)
        return snapshot._from_db_object(self._context,
                                        snapshot, snapshot_db)

    @staticmethod
    def _is_cleanable(status, obj_version):
        # Before 1.6 we didn't have workers table, so cleanup wasn't supported.
        # cleaning.
        if obj_version and obj_version < 1.6:
            return False
        return status in ('creating', 'deleting', 'uploading', 'downloading')

    def begin_attach(self, attach_mode):
        attachment = objects.VolumeAttachment(
            context=self._context,
            attach_status=c_fields.VolumeAttachStatus.ATTACHING,
            volume_id=self.id)
        attachment.create()
        with self.obj_as_admin():
            self.admin_metadata['attached_mode'] = attach_mode
            self.save()
        return attachment

    def finish_detach(self, attachment_id):
        with self.obj_as_admin():
            volume_updates, attachment_updates = (
                db.volume_detached(self._context, self.id, attachment_id))
            db.volume_admin_metadata_delete(self._context, self.id,
                                            'attached_mode')
            self.admin_metadata.pop('attached_mode', None)
        # Remove attachment in volume only when this field is loaded.
        if attachment_updates and self.obj_attr_is_set('volume_attachment'):
            for i, attachment in enumerate(self.volume_attachment):
                if attachment.id == attachment_id:
                    del self.volume_attachment.objects[i]
                    break

        self.update(volume_updates)
        self.obj_reset_changes(
            list(volume_updates.keys()) +
            ['volume_attachment', 'admin_metadata'])

    def is_replicated(self):
        return self.volume_type and self.volume_type.is_replicated()

    def is_multiattach(self):
        return self.volume_type and self.volume_type.is_multiattach()

    # Don't add it as a property to avoid having to add it obj_extra_fields,
    # to manager's _VOLUME_CLONE_SKIP_PROPERTIES, etc.
    def is_migration_target(self):
        return (self.migration_status or '').startswith('target:')


@base.CinderObjectRegistry.register
class VolumeList(base.ObjectListBase, base.CinderObject):
    VERSION = '1.1'

    fields = {
        'objects': fields.ListOfObjectsField('Volume'),
    }

    @staticmethod
    def include_in_cluster(context, cluster, partial_rename=True, **filters):
        """Include all volumes matching the filters into a cluster.

        When partial_rename is set we will not set the cluster_name with
        cluster parameter value directly, we'll replace provided cluster_name
        or host filter value with cluster instead.

        This is useful when we want to replace just the cluster name but leave
        the backend and pool information as it is.  If we are using
        cluster_name to filter, we'll use that same DB field to replace the
        cluster value and leave the rest as it is.  Likewise if we use the host
        to filter.

        Returns the number of volumes that have been changed.
        """
        return db.volume_include_in_cluster(context, cluster, partial_rename,
                                            **filters)

    @classmethod
    def _get_expected_attrs(cls, context, *args, **kwargs):
        expected_attrs = ['metadata', 'volume_type', 'volume_attachment']
        if context.is_admin:
            expected_attrs.append('admin_metadata')

        return expected_attrs

    @classmethod
    def get_all(cls, context, marker=None, limit=None, sort_keys=None,
                sort_dirs=None, filters=None, offset=None):
        volumes = db.volume_get_all(context, marker, limit,
                                    sort_keys=sort_keys, sort_dirs=sort_dirs,
                                    filters=filters, offset=offset)
        expected_attrs = cls._get_expected_attrs(context)
        return base.obj_make_list(context, cls(context), objects.Volume,
                                  volumes, expected_attrs=expected_attrs)

    @classmethod
    def get_all_by_host(cls, context, host, filters=None):
        volumes = db.volume_get_all_by_host(context, host, filters)
        expected_attrs = cls._get_expected_attrs(context)
        return base.obj_make_list(context, cls(context), objects.Volume,
                                  volumes, expected_attrs=expected_attrs)

    @classmethod
    def get_all_by_group(cls, context, group_id, filters=None):
        # Consistency group
        volumes = db.volume_get_all_by_group(context, group_id, filters)
        expected_attrs = cls._get_expected_attrs(context)
        return base.obj_make_list(context, cls(context), objects.Volume,
                                  volumes, expected_attrs=expected_attrs)

    @classmethod
    def get_all_by_generic_group(cls, context, group_id, filters=None):
        # Generic volume group
        volumes = db.volume_get_all_by_generic_group(context, group_id,
                                                     filters)
        expected_attrs = cls._get_expected_attrs(context)
        return base.obj_make_list(context, cls(context), objects.Volume,
                                  volumes, expected_attrs=expected_attrs)

    @classmethod
    def get_all_by_project(cls, context, project_id, marker=None, limit=None,
                           sort_keys=None, sort_dirs=None, filters=None,
                           offset=None):
        volumes = db.volume_get_all_by_project(context, project_id, marker,
                                               limit, sort_keys=sort_keys,
                                               sort_dirs=sort_dirs,
                                               filters=filters, offset=offset)
        expected_attrs = cls._get_expected_attrs(context)
        return base.obj_make_list(context, cls(context), objects.Volume,
                                  volumes, expected_attrs=expected_attrs)

    @classmethod
    def get_volume_summary(cls, context, project_only, filters=None):
        volumes = db.get_volume_summary(context, project_only, filters)
        return volumes

    @classmethod
    def get_all_active_by_window(cls, context, begin, end):
        volumes = db.volume_get_all_active_by_window(context, begin, end)
        expected_attrs = cls._get_expected_attrs(context)
        return base.obj_make_list(context, cls(context), objects.Volume,
                                  volumes, expected_attrs=expected_attrs)



# Copyright (c) 2011-2012 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Filter support
"""
from typing import Iterable

from oslo_log import log as logging

from cinder.scheduler import base_handler

LOG = logging.getLogger(__name__)


class BaseFilter(object):
    """Base class for all filter classes."""
    def _filter_one(self, obj, filter_properties) -> bool:
        """Return True if it passes the filter, False otherwise.

        Override this in a subclass.
        """
        return True

    def filter_all(self, filter_obj_list, filter_properties):
        """Yield objects that pass the filter.

        Can be overridden in a subclass, if you need to base filtering
        decisions on all objects.  Otherwise, one can just override
        _filter_one() to filter a single object.
        """
        for obj in filter_obj_list:
            if self._filter_one(obj, filter_properties):
                yield obj

    # Set to true in a subclass if a filter only needs to be run once
    # for each request rather than for each instance
    run_filter_once_per_request = False

    def run_filter_for_index(self, index):
        """Return True if the filter needs to be run for n-th instances.

        Only need to override this if a filter needs anything other than
        "first only" or "all" behaviour.
        """
        return not (self.run_filter_once_per_request and index > 0)


class BaseFilterHandler(base_handler.BaseHandler):
    """Base class to handle loading filter classes.

    This class should be subclassed where one needs to use filters.
    """

    def _log_filtration(self, full_filter_results,
                        part_filter_results, filter_properties):
        # Log the filtration history
        rspec = filter_properties.get("request_spec", {})
        msg_dict = {"vol_id": rspec.get("volume_id", ""),
                    "str_results": full_filter_results}
        LOG.debug("Filtering removed all hosts for the request with "
                  "volume ID '%(vol_id)s'. Filter results: %(str_results)s",
                  msg_dict)
        msg_dict["str_results"] = ', '.join(
            "%(cls_name)s: (start: %(start)s, end: %(end)s)" % {
                "cls_name": value[0], "start": value[1], "end": value[2]}
            for value in part_filter_results)
        LOG.info("Filtering removed all hosts for the request with "
                 "volume ID '%(vol_id)s'. Filter results: %(str_results)s",
                 msg_dict)

    def get_filtered_objects(self, filter_classes, objs: Iterable,
                             filter_properties: dict, index: int = 0) -> list:
        """Get objects after filter

        :param filter_classes: filters that will be used to filter the
                               objects
        :param objs: objects that will be filtered
        :param filter_properties: client filter properties
        :param index: This value needs to be increased in the caller
                      function of get_filtered_objects when handling
                      each resource.
        """
        list_objs = list(objs)
        LOG.debug("Starting with %d host(s)", len(list_objs))
        # The 'part_filter_results' list just tracks the number of hosts
        # before and after the filter, unless the filter returns zero
        # hosts, in which it records the host/nodename for the last batch
        # that was removed. Since the full_filter_results can be very large,
        # it is only recorded if the LOG level is set to debug.
        part_filter_results = []
        full_filter_results = []
        for filter_cls in filter_classes:
            cls_name = filter_cls.__name__
            start_count = len(list_objs)
            filter_class = filter_cls()

            if filter_class.run_filter_for_index(index):
                objs = filter_class.filter_all(list_objs, filter_properties)
                if objs is None:
                    LOG.info("Filter %s returned 0 hosts", cls_name)
                    full_filter_results.append((cls_name, None))
                    list_objs = None
                    break

                list_objs = list(objs)
                end_count = len(list_objs)
                part_filter_results.append((cls_name, start_count, end_count))
                remaining = [getattr(obj, "host", obj)
                             for obj in list_objs]
                full_filter_results.append((cls_name, remaining))

                LOG.debug("Filter %(cls_name)s returned "
                          "%(obj_len)d host(s)",
                          {'cls_name': cls_name, 'obj_len': len(list_objs)})
        if not list_objs:
            self._log_filtration(full_filter_results,
                                 part_filter_results, filter_properties)
        return list_objs



# Copyright (C) 2012 - 2014 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from cinder.api import common


class ViewBuilder(common.ViewBuilder):
    """Model cgsnapshot API responses as a python dictionary."""

    _collection_name = "cgsnapshots"

    def __init__(self):
        """Initialize view builder."""
        super(ViewBuilder, self).__init__()

    def summary_list(self, request, cgsnapshots):
        """Show a list of cgsnapshots without many details."""
        return self._list_view(self.summary, request, cgsnapshots)

    def detail_list(self, request, cgsnapshots):
        """Detailed view of a list of cgsnapshots ."""
        return self._list_view(self.detail, request, cgsnapshots)

    def summary(self, request, cgsnapshot):
        """Generic, non-detailed view of a cgsnapshot."""
        return {
            'cgsnapshot': {
                'id': cgsnapshot.id,
                'name': cgsnapshot.name
            }
        }

    def detail(self, request, cgsnapshot):
        """Detailed view of a single cgsnapshot."""
        try:
            group_id = cgsnapshot.consistencygroup_id
        except AttributeError:
            try:
                group_id = cgsnapshot.group_id
            except AttributeError:
                group_id = None
        else:
            group_id = None

        return {
            'cgsnapshot': {
                'id': cgsnapshot.id,
                'consistencygroup_id': group_id,
                'status': cgsnapshot.status,
                'created_at': cgsnapshot.created_at,
                'name': cgsnapshot.name,
                'description': cgsnapshot.description
            }
        }

    def _list_view(self, func, request, cgsnapshots):
        """Provide a view for a list of cgsnapshots."""
        cgsnapshots_list = [func(request, cgsnapshot)['cgsnapshot']
                            for cgsnapshot in cgsnapshots]
        cgsnapshots_dict = dict(cgsnapshots=cgsnapshots_list)

        return cgsnapshots_dict



# Copyright 2020 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


class ViewBuilder(object):
    """Model default type API response as a python dictionary."""

    _collection_name = "default_types"

    def _convert_to_dict(self, default):
        return {'project_id': default.project_id,
                'volume_type_id': default.volume_type_id}

    def create(self, default_type):
        """Detailed view of a default type when set."""

        return {'default_type': self._convert_to_dict(default_type)}

    def index(self, default_types):
        """Build a view of a list of default types.

         .. code-block:: json

             {"default_types":
                [
                 {
                   "project_id": "248592b4-a6da-4c4c-abe0-9d8dbe0b74b4",
                   "volume_type_id": "7152eb1e-aef0-4bcd-a3ab-46b7ef17e2e6"
                 },
                 {
                   "project_id": "1234567-4c4c-abcd-abe0-1a2b3c4d5e6ff",
                   "volume_type_id": "5e3b298a-f1fc-4d32-9828-0d720da81ddd"
                 }
             ]
             }
        """

        default_types_view = []
        for default_type in default_types:
            default_types_view.append(self._convert_to_dict(default_type))

        return {'default_types': default_types_view}

    def detail(self, default_type):
        """Build a view of a default type.

         .. code-block:: json

             {"default_type":
                 {
                   "project_id": "248592b4-a6da-4c4c-abe0-9d8dbe0b74b4",
                   "volume_type_id": "6bd1de9a-b8b5-4c43-a597-00170ab06b50"
                 }
             }
        """
        return {'default_type': self._convert_to_dict(default_type)}



# Copyright (c) 2016 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import enum

import six


class Enum(enum.Enum):
    @classmethod
    def verify(cls, value, allow_none=True):
        if value is None and not allow_none:
            raise ValueError(
                'None is not allowed here for %s.') % cls.__name__
        elif value is not None and not isinstance(value, cls):
            raise ValueError('%(value)s is not an instance of %(name)s.') % {
                'value': value, 'name': cls.__name__}

    @classmethod
    def get_all(cls):
        return list(cls)

    @classmethod
    def get_opt(cls, value):
        option_map = cls.get_option_map()
        if option_map is None:
            raise NotImplementedError(
                'Option map is not defined for %s.') % cls.__name__

        ret = option_map.get(value, None)
        if ret is None:
            raise ValueError('%(value)s is not a valid option for %(name)s.'
                             ) % {'value': value, 'name': cls.__name__}
        return ret

    @classmethod
    def parse(cls, value):
        if isinstance(value, six.string_types):
            ret = cls.from_str(value)
        elif isinstance(value, six.integer_types):
            ret = cls.from_int(value)
        elif isinstance(value, cls):
            ret = value
        elif value is None:
            ret = None
        else:
            raise ValueError(
                'Not supported value type: %s.') % type(value)
        return ret

    def is_equal(self, value):
        if isinstance(value, six.string_types):
            ret = self.value.lower() == value.lower()
        else:
            ret = self.value == value
        return ret

    @classmethod
    def from_int(cls, value):
        ret = None
        int_index = cls.get_int_index()
        if int_index is not None:
            try:
                ret = int_index[value]
            except IndexError:
                pass
        else:
            try:
                ret = next(i for i in cls.get_all() if i.is_equal(value))
            except StopIteration:
                pass
        if ret is None:
            raise ValueError
        return ret

    @classmethod
    def from_str(cls, value):
        ret = None
        if value is not None:
            for item in cls.get_all():
                if item.is_equal(value):
                    ret = item
                    break
            else:
                cls._raise_invalid_value(value)
        return ret

    @classmethod
    def _raise_invalid_value(cls, value):
        msg = ('%(value)s is not a valid value for %(name)s.'
               ) % {'value': value, 'name': cls.__name__}
        raise ValueError(msg)

    @classmethod
    def get_option_map(cls):
        raise None

    @classmethod
    def get_int_index(cls):
        return None

    @classmethod
    def values(cls):
        return [m.value for m in cls.__members__.values()]

    @classmethod
    def enum_name(cls):
        return cls.__name__


class VNXCtrlMethod(object):
    LIMIT_CTRL = 'limit'

    def __init__(self, method, metric, value):
        self.method = method
        self.metric = metric
        self.value = value



# Copyright 2016 Dell Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""
Backup driver with 'chunked' backup operations.
"""

from cinder.interface import backup_driver


class BackupChunkedDriver(backup_driver.BackupDriver):
    """Backup driver that supports 'chunked' backups."""

    def put_container(self, container):
        """Create the container if needed. No failure if it pre-exists.

        :param container: The container to write into.
        """

    def get_container_entries(self, container, prefix):
        """Get container entry names.

        :param container: The container from which to get entries.
        :param prefix: The prefix used to match entries.
        """

    def get_object_writer(self, container, object_name, extra_metadata=None):
        """Returns a writer which stores the chunk data in backup repository.

       :param container: The container to write to.
       :param object_name: The object name to write.
       :param extra_metadata: Extra metadata to be included.
       :returns: A context handler that can be used in a "with" context.
        """

    def get_object_reader(self, container, object_name, extra_metadata=None):
        """Returns a reader object for the backed up chunk.

       :param container: The container to read from.
       :param object_name: The object name to read.
       :param extra_metadata: Extra metadata to be included.
       """

    def delete_object(self, container, object_name):
        """Delete object from container.

       :param container: The container to modify.
       :param object_name: The object name to delete.
       """

    def update_container_name(self, backup, container):
        """Allows sub-classes to override container name.

        This method exists so that sub-classes can override the container name
        as it comes in to the driver in the backup object. Implementations
        should return None if no change to the container name is desired.
        """

    def get_extra_metadata(self, backup, volume):
        """Return extra metadata to use in prepare_backup.

        This method allows for collection of extra metadata in prepare_backup()
        which will be passed to get_object_reader() and get_object_writer().
        Subclass extensions can use this extra information to optimize
        data transfers.

         :returns: json serializable object
        """
